##  Module histogram.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

import collections
import random
import threading
import abc
import time
import operator
import math

from . import statistics, exceptions, py3comp


DEFAULT_UNIFORM_RESERVOIR_SIZE = 1028
DEFAULT_TIME_WINDOW_SIZE = 60
DEFAULT_EXPONENTIAL_DECAY_FACTOR = 0.015


def search_greater(values, target):
    """
    Return the first index for which target is greater or equal to the first
    item of the tuple found in values
    """
    first = 0
    last = len(values)

    while first < last:
        middle = (first + last) // 2
        if values[middle][0] < target:
            first = middle + 1
        else:
            last = middle

    return first


class ReservoirBase(object):
    __metaclass__ = abc.ABCMeta

    """
    Base class for reservoirs. Subclass and override _do_add, _get_values and
    _same_parameters
    """

    def add(self, value):
        """
        Add a value to the reservoir
        The value will be casted to a floating-point, so a TypeError or a
        ValueError may be raised.
        """

        if not isinstance(value, float):
            value = float(value)

        return self._do_add(value)

    @property
    def values(self):
        """
        Return the stored values
        """

        return self._get_values()

    @property
    def sorted_values(self):
        """
        Sort and return the current sample values
        """

        return sorted(self.values)

    def same_kind(self, other):
        """
        Return True if "other" is an object of the same type and it was
        instantiated with the same parameters
        """

        return type(self) is type(other) and self._same_parameters(other)

    @abc.abstractmethod
    def _do_add(self, value):
        """
        Add the floating-point value to the reservoir. Override in subclasses
        """

    @abc.abstractmethod
    def _get_values(self):
        """
        Get the current reservoir's content. Override in subclasses
        """

    @abc.abstractmethod
    def _same_parameters(self, other):
        """
        Return True if this object has been instantiated with the same
        parameters as "other".
        Override in subclasses
        """


class UniformReservoir(ReservoirBase):
    """
    A random sampling reservoir of floating-point values. Uses Vitter's
    Algorithm R to produce a statistically representative sample
    (http://www.cs.umd.edu/~samir/498/vitter.pdf)
    """

    def __init__(self, size=DEFAULT_UNIFORM_RESERVOIR_SIZE):
        self.size = size
        self._values = [0] * size
        self.count = 0
        self.lock = threading.Lock()

    def _do_add(self, value):
        changed = False

        with self.lock:
            if self.count < self.size:
                self._values[self.count] = value
                changed = True
            else:
                # not randint() because it yields different values on
                # python 3, it would be a nightmare to test.
                k = int(random.uniform(0, self.count))
                if k < self.size:
                    self._values[k] = value
                    changed = True

            self.count += 1

        return changed

    def _get_values(self):
        return self._values[:min(self.count, self.size)]

    def _same_parameters(self, other):
        return self.size == other.size

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.size)


class SlidingWindowReservoir(ReservoirBase):
    """
    A simple sliding-window reservoir that keeps the last N values
    """

    def __init__(self, size=DEFAULT_UNIFORM_RESERVOIR_SIZE):
        self.size = size
        self.deque = collections.deque(maxlen=self.size)

    def _do_add(self, value):
        # No need for explicit lock - deques should be thread-safe:
        # http://docs.python.org/2/library/collections.html#collections.deque
        self.deque.append(value)

    def _get_values(self):
        return list(self.deque)

    def _same_parameters(self, other):
        return self.size == other.size

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.size)


class SlidingTimeWindowReservoir(ReservoirBase):
    """
    A time-sliced reservoir that keeps the values added in the last N seconds
    """

    def __init__(self, window_size=DEFAULT_TIME_WINDOW_SIZE):
        """
        Build a new sliding time-window reservoir
        window_size is the time window size in seconds
        """
        self.window_size = window_size
        self.lock = threading.Lock()
        self.key = operator.itemgetter(0)
        self._values = []

    def _do_add(self, value):
        now = time.time()

        with self.lock:
            self.tick(now)

        self._values.append((now, value))

    def tick(self, now):
        target = now - self.window_size

        # the values are sorted by the first element (timestamp), so let's
        # perform a dichotomic search
        idx = search_greater(self._values, target)

        # older values found, discard them
        if idx:
            self._values = self._values[idx:]

    def _get_values(self):
        now = time.time()

        with self.lock:
            self.tick(now)

        return [y for x, y in self._values]

    def _same_parameters(self, other):
        return self.window_size == other.window_size

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.window_size)


class ExponentialDecayingReservoir(ReservoirBase):
    """
    An exponential-weighted reservoir which exponentially decays older values
    in order to give greater significance to newer ones.
    See http://dimacs.rutgers.edu/~graham/pubs/papers/fwddecay.pdf
    """

    # TODO: replace the sort()s with a proper data structure (btree/skiplist).
    # However, since the list is keep sorted (and it should be very small),
    # the sort() shouldn't dramatically slow down the insertions, also
    # considering that the search can be log(n) in that way

    RESCALE_THRESHOLD = 3600
    EPSILON = 1e-12

    def __init__(self, size=DEFAULT_UNIFORM_RESERVOIR_SIZE,
                 alpha=DEFAULT_EXPONENTIAL_DECAY_FACTOR):
        self.size = size
        self.alpha = alpha
        self.start_time = time.time()
        self.lock = threading.Lock()
        self.count = 0
        self.next_scale_time = self.start_time + self.RESCALE_THRESHOLD
        self.key = operator.itemgetter(0)

        self._values = []

    def _lookup(self, timestamp):
        """
        Return the index of the value associated with "timestamp" if any, else
        None. Since the timestamps are floating-point values, they are
        considered equal if their absolute difference is smaller than
        self.EPSILON
        """

        idx = search_greater(self._values, timestamp)
        if (idx < len(self._values)
                and math.fabs(self._values[idx][0] - timestamp) < self.EPSILON):
            return idx
        return None

    def _put(self, timestamp, value):
        """Replace the value associated with "timestamp" or add the new value"""

        idx = self._lookup(timestamp)
        if idx is not None:
            self._values[idx] = (timestamp, value)
        else:
            self._values.append((timestamp, value))

    def _do_add(self, value):
        now = time.time()

        self.rescale(now)

        rnd = random.random()
        weighted_time = self.weight(now - self.start_time) / rnd

        changed = False

        with self.lock:
            if self.count < self.size:
                self._put(weighted_time, value)
                self._values.sort(key=self.key)
                changed = True
            else:
                first = self._values[0][0]
                if first < weighted_time:
                    idx = self._lookup(weighted_time)
                    if idx is None:
                        self._values[0] = (weighted_time, value)
                        self._values.sort(key=self.key)
                        changed = True

            self.count += 1

        return changed

    def weight(self, t):
        return math.exp(self.alpha * t)

    def rescale(self, now):
        with self.lock:
            if now > self.next_scale_time:
                original_values = self._values[:]
                self._values = []
                for i, (k, v) in enumerate(original_values):
                    k *= math.exp(-self.alpha * (now - self.start_time))
                    self._put(k, v)

                self.count = len(self._values)
                self.start_time = now
                self.next_scale_time = self.start_time + self.RESCALE_THRESHOLD

    def _get_values(self):
        return [y for x, y in self._values[:max(self.count, self.size)]]

    def _same_parameters(self, other):
        return self.size == other.size and self.alpha == other.alpha

    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.size, self.alpha)


class Histogram(object):
    """A metric which calculates some statistics over the distribution of some
    values"""

    def __init__(self, reservoir):
        self.reservoir = reservoir

    def notify(self, value):
        """Add a new value to the metric"""

        return self.reservoir.add(value)

    def raw_data(self):
        """Return the raw underlying data"""

        return self.reservoir.values

    def get(self):
        """Return the computed statistics over the gathered data"""

        values = self.reservoir.sorted_values

        def safe(f, *args):
            try:
                return f(values, *args)
            except exceptions.StatisticsError:
                return 0.0

        plevels = [50, 75, 90, 95, 99, 99.9]
        percentiles = [safe(statistics.percentile, p) for p in plevels]

        try:
            histogram = statistics.get_histogram(values)
        except exceptions.StatisticsError:
            histogram = [(0, 0)]

        res = dict(
            kind="histogram",
            min=values[0] if values else 0,
            max=values[-1] if values else 0,
            arithmetic_mean=safe(statistics.mean),
            geometric_mean=safe(statistics.geometric_mean),
            harmonic_mean=safe(statistics.harmonic_mean),
            median=safe(statistics.median),
            variance=safe(statistics.variance),
            standard_deviation=safe(statistics.stdev),
            skewness=safe(statistics.skewness),
            kurtosis=safe(statistics.kurtosis),
            percentile=py3comp.zip(plevels, percentiles),
            histogram=histogram,
            n=len(values))
        return res




from nose.tools import assert_equal

from .. import simple_metrics as mm


class TestCounter(object):
    def setUp(self):
        self.obj = mm.Counter()

    def test_notify(self):
        self.obj.notify(5)
        assert_equal(self.obj.value, 5)

        self.obj.notify(-5)
        assert_equal(self.obj.value, 0)

        self.obj.notify(-1)
        assert_equal(self.obj.value, -1)

    def test_notify_not_integer(self):
        self.obj.notify(1.4)
        assert_equal(self.obj.value, 1)

    def test_get(self):
        self.obj.value = 3
        assert_equal(self.obj.get(), dict(kind="counter", value=3))

    def test_raw_data(self):
        self.obj.value = 3
        assert_equal(self.obj.raw_data(), 3)


class TestGauge(object):
    def setUp(self):
        self.obj = mm.Gauge()

    def test_notify(self):
        assert_equal(self.obj.value, None)
        self.obj.notify("version 1.0")
        assert_equal(self.obj.value, "version 1.0")

        self.obj.notify(1.23)
        assert_equal(self.obj.value, 1.23)

    def test_get(self):
        self.obj.value = "test"
        assert_equal(self.obj.get(), dict(kind="gauge", value="test"))

    def test_raw_data(self):
        self.obj.value = "test"
        assert_equal(self.obj.raw_data(), "test")



##  Module meter.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.


"""
Meter metric - measures throughput at various time intervals
"""

import math
import time
import threading


DEFAULT_TICK_INTERVAL = 5


class EWMA(object):
    """
    Compute exponential-weighted moving average of values incoming at a fixed rate.

    http://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average
    """

    def __init__(self, time_period, tick_interval):
        """
        time_unit is the period of time on which the moving average is computed,
        expressed in minutes
        """

        self.time_period = time_period
        self.tick_interval = tick_interval

        self.rate = 0.0
        self.value = 0
        self.initialized = False

        self.alpha = self.compute_alpha(time_period, tick_interval)

        self.lock = threading.Lock()

    @staticmethod
    def compute_alpha(period, interval):
        """Compute exponential smoothing factor"""

        return 1 - math.exp(-interval / (60.0 * period))

    def update(self, value):
        """
        Update the current rate with the given value.
        The value must be an integer.
        """

        value = int(value)

        with self.lock:
            self.value += value

    def tick(self):
        """Decay the current rate according to the elapsed time"""

        instant_rate = float(self.value) / float(self.tick_interval)

        with self.lock:
            if self.initialized:
                self.rate += (self.alpha * (instant_rate - self.rate))
            else:
                self.initialized = True
                self.rate = instant_rate

            self.value = 0


class Meter(object):
    """
    A meter metric which measures mean throughput and one, five, and fifteen-minute
    exponentially-weighted moving average throughputs.
    This is very similar to the unix "load average" metric. All the throughput
    values are expressed in number of operation per second.
    """

    def __init__(self, tick_interval=DEFAULT_TICK_INTERVAL):
        self.tick_interval = tick_interval

        # one minute
        self.m1 = EWMA(1, tick_interval)
        # five minutes
        self.m5 = EWMA(5, tick_interval)
        # fifteen minutes
        self.m15 = EWMA(15, tick_interval)
        # one day
        self.day = EWMA(60 * 24, tick_interval)

        self.started_on = self.latest_tick = time.time()

        self.count = 0

        self.lock = threading.Lock()

    def notify(self, value):
        """Add a new observation to the metric"""

        with self.lock:
            #TODO: this could slow down slow-rate incoming updates
            # since the number of ticks depends on the actual time
            # passed since the latest notification. Consider using
            # a real timer to tick the EWMA.

            self.tick()

            for avg in (self.m1, self.m5, self.m15, self.day):
                avg.update(value)
            self.count += value

    def tick_all(self, times):
        """
        Tick all the EWMAs for the given number of times
        """

        for i in range(times):
            for avg in (self.m1, self.m5, self.m15, self.day):
                avg.tick()

    def tick(self):
        """
        Emulate a timer: in order to avoid a real timer we "tick" a number
        of times depending on the actual time passed since the last tick
        """

        now = time.time()

        elapsed = now - self.latest_tick

        if elapsed > self.tick_interval:
            ticks = int(elapsed / self.tick_interval)

            self.tick_all(ticks)

            self.latest_tick = now

    def raw_data(self):
        """Return the raw underlying data"""

        return self.count

    def get(self):
        """
        Return the computed statistics over the gathered data
        """

        with self.lock:
            self.tick()

            data = dict(
                kind="meter",
                count=self.count,
                mean=self.count / (time.time() - self.started_on),
                one=self.m1.rate,
                five=self.m5.rate,
                fifteen=self.m15.rate,
                day=self.day.rate)

        return data

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.tick_interval)



##  Module reporter.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

"""
Reporting facilities
"""


import os
import csv
import logging
import uuid
import threading
import time
import atexit

from . import metrics, py3comp


log = logging.getLogger('appmetrics.reporter')


REGISTRY = {}
LOCK = threading.Lock()


def register(callback, schedule, tag=None):
    """
    Register a callback which will be called at scheduled intervals with
    the metrics that have the given tag (or all the metrics if None).
    Return an identifier which can be used to access the registered callback later.
    """

    try:
        iter(schedule)
    except TypeError:
        raise TypeError("{} is not iterable".format(schedule))

    if not callable(callback):
        raise TypeError("{} is not callable".format(callback))

    thread = Timer(schedule, callback, tag)

    id_ = str(uuid.uuid4())
    with LOCK:
        REGISTRY[id_] = thread

    thread.start()

    return id_


def get(id_):
    """
    Return the registered callback with the given name, or None if it does not exist
    """
    return REGISTRY.get(id_)


def remove(id_):
    """
    Remove the callback and its schedule
    """
    with LOCK:
        thread = REGISTRY.pop(id_, None)
        if thread is not None:
            thread.cancel()

    return thread


class Timer(threading.Thread):
    """
    Encapsulate a callback and its parameters
    """

    def __init__(self, schedule, callback, tag=None):
        """
        "schedule" must be an iterator yielding the next "tick" at each iteration
        """
        super(Timer, self).__init__()

        self.schedule = schedule
        self.callback = callback
        self.tag = tag
        self._event = threading.Event()
        self.daemon = True

    def run(self):
        while not self._event.is_set():
            now = time.time()
            # skip already passed ticks
            for next_time in self.schedule:
                if next_time > now:
                    break
            # the iterator was consumed, exit
            else:
                break

            # sleep for the remaining time
            time.sleep(next_time - now)

            # the event may have been set while sleeping
            if not self._event.is_set():
                data = get_metrics(self.tag)

                if not data:
                    log.debug("No metrics found for tag: {}".format(self.tag))
                else:
                    # call the function, finally
                    self.callback(data)

        # set the event, for consistency
        self.cancel()

    def cancel(self):
        """
        Cancel the timer, no more actions will be performed
        """
        self._event.set()

    @property
    def is_running(self):
        return not self._event.is_set()


def get_metrics(tag):
    """
    Return the values for the metrics with the given tag or all the available metrics if None
    """
    if tag is None:
        return metrics.metrics_by_name_list(metrics.metrics())
    else:
        return metrics.metrics_by_tag(tag)


def fixed_interval_scheduler(interval):
    """
    A scheduler that ticks at fixed intervals of "interval" seconds
    """
    start = time.time()
    next_tick = start

    while True:
        next_tick += interval
        yield next_tick


class CSVReporter(object):
    histogram_header = (
        'time', 'n', 'min', 'max', 'arithmetic_mean', 'median', 'harmonic_mean', 'geometric_mean',
        'standard_deviation', 'variance', 'percentile_50', 'percentile_75', 'percentile_90',
        'percentile_95', 'percentile_99', 'percentile_99.9', 'kurtosis', 'skewness')

    meter_header = ('time', 'count', 'mean', 'one', 'five', 'fifteen', 'day')

    def __init__(self, directory):
        self.directory = directory

    def file_name(self, name, kind):
        file_name = os.path.join(self.directory, "{}_{}.csv".format(name, kind))

        if not os.path.exists(file_name):
            new = True
        else:
            new = False

        return file_name, new

    def dump_histogram(self, name, obj):

        # histogram doesn't fit into a tabular format
        obj.pop('histogram')

        # we already know its kind
        kind = obj.pop('kind')

        # flatten percentiles
        percentiles = obj.pop('percentile')
        for k, v in percentiles:
            obj['percentile_{}'.format(k)] = v

        # add the current time
        obj['time'] = time.time()

        file_name, new = self.file_name(name, kind)

        with open(file_name, "a" if py3comp.PY3 else "ab") as of:
            writer = csv.DictWriter(of, self.histogram_header)

            # if the file is new, write the header once
            if new:
                writer.writerow(dict(zip(self.histogram_header, self.histogram_header)))

            writer.writerow(obj)

        return file_name

    def dump_meter(self, name, obj):

        # we already know its kind
        kind = obj.pop('kind')

        # add the current time
        obj['time'] = time.time()

        file_name, new = self.file_name(name, kind)

        with open(file_name, "a" if py3comp.PY3 else "ab") as of:
            writer = csv.DictWriter(of, self.meter_header)

            # if the file is new, write the header once
            if new:
                writer.writerow(dict(zip(self.meter_header, self.meter_header)))

            writer.writerow(obj)

        return file_name

    def __call__(self, objects):
        for name, obj in py3comp.iteritems(objects):
            fun = getattr(self, "dump_%s" % obj.get('kind', "unknown"), None)
            if fun:
                # protect the original object
                fun(name, obj.copy())


@atexit.register
def cleanup():
    for v in REGISTRY.values():
        v.cancel()



import random

from nose import tools as nt
import mock

from .. import histogram as mm
from ..py3comp import assert_items_equal

def test_uniform_reservoir_defaults():
    ur = mm.UniformReservoir()
    nt.assert_equal(ur.size, mm.DEFAULT_UNIFORM_RESERVOIR_SIZE)
    nt.assert_equal(ur._values, [0] * mm.DEFAULT_UNIFORM_RESERVOIR_SIZE)
    nt.assert_equal(ur.values, [])
    nt.assert_equal(ur.sorted_values, [])
    nt.assert_equal(ur.count, 0)


def test_search_greater():
    values = [(1, "a"), (2, "b"), (3, "c"), (4, "d"), (5, "e")]

    tests = [
        (0, 0),
        (1, 0),
        (1.5, 1),
        (2, 1),
        (4.5, 4),
        (5, 4),
        (6, 5)]

    for target, expected in tests:
        f = lambda target, expected: nt.assert_equal(expected, mm.search_greater(values, target))
        yield f, target, expected


class TestUniformReservoir(object):
    def setUp(self):
        self.state = random.getstate()
        random.seed(42)

        self.size = 5
        self.ur = mm.UniformReservoir(self.size)

    def tearDown(self):
        random.setstate(self.state)

    def test_add_first(self):
        self.ur.add(1.5)
        nt.assert_equal(self.ur.values, [1.5])

        self.ur.add(2.5)
        nt.assert_equal(self.ur.values, [1.5, 2.5])

        self.ur.add(3.5)
        nt.assert_equal(self.ur.values, [1.5, 2.5, 3.5])

        self.ur.add(4.5)
        nt.assert_equal(self.ur.values, [1.5, 2.5, 3.5, 4.5])

        self.ur.add(5.5)
        nt.assert_equal(self.ur.values, [1.5, 2.5, 3.5, 4.5, 5.5])

    def test_add_overflow(self):
        for i in range(5):
            self.ur.add(i + 1.5)

        nt.assert_equal(self.ur.values, [1.5, 2.5, 3.5, 4.5, 5.5])

        self.ur.add(10)
        nt.assert_equal(self.ur.values, [1.5, 2.5, 3.5, 10.0, 5.5])

        self.ur.add(11)
        nt.assert_equal(self.ur.values, [11, 2.5, 3.5, 10.0, 5.5])

        self.ur.add(12)
        nt.assert_equal(self.ur.values, [11, 12, 3.5, 10.0, 5.5])

        self.ur.add(13)
        nt.assert_equal(self.ur.values, [11, 13, 3.5, 10.0, 5.5])

        self.ur.add(14)
        nt.assert_equal(self.ur.values, [11, 13, 3.5, 10.0, 5.5])

        self.ur.add(15)
        nt.assert_equal(self.ur.values, [11, 13, 3.5, 10.0, 5.5])

    def test_values_smaller_count(self):
        for i in range(3):
            self.ur.add(i)
        nt.assert_equal(len(self.ur.values), self.ur.count)

    def test_values_greater_count(self):
        for i in range(6):
            self.ur.add(i)
        nt.assert_equal(len(self.ur.values), self.ur.size)

    def test_sorted_values(self):
        for i in range(5):
            self.ur.add(random.randint(1, 10))

        random.seed(42)
        nt.assert_equal(self.ur.sorted_values, sorted(random.randint(1, 10) for i in range(5)))

    @nt.raises(TypeError)
    def test_add_bad_type(self):
        self.ur.add(None)

    def test_same_kind(self):
        other = mm.UniformReservoir(self.ur.size)
        nt.assert_true(self.ur.same_kind(other))

    def test_same_kind_with_different_class(self):
        other = mm.SlidingWindowReservoir(self.ur.size)
        nt.assert_false(self.ur.same_kind(other))

    def test_same_kind_with_different_parameters(self):
        other = mm.UniformReservoir(10)
        nt.assert_false(self.ur.same_kind(other))


class TestSlidingWindowReservoir(object):
    def setUp(self):
        self.state = random.getstate()
        random.seed(42)

        self.size = 5
        self.swr = mm.SlidingWindowReservoir(self.size)

    def tearDown(self):
        random.setstate(self.state)

    def test_add_first(self):
        self.swr.add(1.5)
        nt.assert_equal(self.swr.values, [1.5])

        self.swr.add(2.5)
        nt.assert_equal(self.swr.values, [1.5, 2.5])

        self.swr.add(3.5)
        nt.assert_equal(self.swr.values, [1.5, 2.5, 3.5])

        self.swr.add(4.5)
        nt.assert_equal(self.swr.values, [1.5, 2.5, 3.5, 4.5])

        self.swr.add(5.5)
        nt.assert_equal(self.swr.values, [1.5, 2.5, 3.5, 4.5, 5.5])

    def test_add_overflow(self):
        for i in range(5):
            self.swr.add(i + 1.5)

        nt.assert_equal(self.swr.values, [1.5, 2.5, 3.5, 4.5, 5.5])

        self.swr.add(10)
        nt.assert_equal(self.swr.values, [2.5, 3.5, 4.5, 5.5, 10.0])

        self.swr.add(11)
        nt.assert_equal(self.swr.values, [3.5, 4.5, 5.5, 10.0, 11.0])

    def test_sorted_values(self):
        for i in range(5):
            self.swr.add(random.randint(1, 10))

        random.seed(42)
        nt.assert_equal(self.swr.sorted_values, sorted(random.randint(1, 10) for i in range(5)))

    @nt.raises(TypeError)
    def test_add_bad_type(self):
        self.swr.add(None)

    def test_same_kind(self):
        other = mm.SlidingWindowReservoir(self.swr.size)
        nt.assert_true(self.swr.same_kind(other))

    def test_same_kind_with_different_class(self):
        other = mm.UniformReservoir(self.swr.size)
        nt.assert_false(self.swr.same_kind(other))

    def test_same_kind_with_different_parameters(self):
        other = mm.SlidingWindowReservoir(10)
        nt.assert_false(self.swr.same_kind(other))


class TestSlidingTimeWindowReservoir(object):
    def setUp(self):
        self.patch = mock.patch('appmetrics.histogram.time.time')
        self.time = self.patch.start()

        self.window_size = 3 # seconds
        self.rr = mm.SlidingTimeWindowReservoir(self.window_size)

    def tearDown(self):
        self.patch.stop()

    @nt.raises(TypeError)
    def test_add_bad_type(self):
        self.rr.add(None)

    def test_add(self):
        self.time.return_value = 1.0

        for i in range(10):
            self.rr.add(i)

        nt.assert_equal(list(self.rr._values), [(1.0, float(x)) for x in range(10)])

    def test_add_exceeded_time(self):
        self.time.return_value = 1
        self.rr.add(1)

        nt.assert_equal(list(self.rr._values), [(1, 1)])

        self.time.return_value = 1.1
        self.rr.add(2)
        nt.assert_equal(list(self.rr._values), [(1, 1), (1.1, 2)])

        self.time.return_value = 1.2
        self.rr.add(3)
        nt.assert_equal(list(self.rr._values), [(1, 1), (1.1, 2), (1.2, 3)])

        self.time.return_value = 1.3
        self.rr.add(4)
        nt.assert_equal(list(self.rr._values), [(1, 1), (1.1, 2), (1.2, 3), (1.3, 4)])

        self.time.return_value = 3.1
        self.rr.add(5)
        nt.assert_equal(list(self.rr._values), [(1, 1), (1.1, 2), (1.2, 3), (1.3, 4), (3.1, 5)])

        self.time.return_value = 4.05
        self.rr.add(6)
        nt.assert_equal(list(self.rr._values), [(1.1, 2), (1.2, 3), (1.3, 4), (3.1, 5), (4.05, 6)])

        self.time.return_value = 4.1
        self.rr.add(7)
        nt.assert_equal(list(self.rr._values), [(1.1, 2), (1.2, 3), (1.3, 4), (3.1, 5), (4.05, 6), (4.1, 7)])

        self.time.return_value = 4.2
        self.rr.add(8)
        nt.assert_equal(list(self.rr._values), [(1.3, 4), (3.1, 5), (4.05, 6), (4.1, 7), (4.2, 8)])

        self.time.return_value = 10
        self.rr.add(9)
        nt.assert_equal(list(self.rr._values), [(10, 9)])

    def test_values(self):
        self.rr._values = [(1, 10), (1.5, 1.5), (2, 2), (3, 3)]
        self.time.return_value = 3.0
        nt.assert_equal(self.rr.values, [10, 1.5, 2, 3])

    def test_values_exceeded_time(self):
        self.rr._values = [(1, 10), (2, 2), (3, 1), (4, 4)]
        self.time.return_value = 4.0001
        nt.assert_equal(self.rr.values, [2, 1, 4])

    def test_sorted_values(self):
        self.rr._values = [(1, 10), (2, 2), (3, 1), (4, 4)]
        self.time.return_value = 4.0001
        nt.assert_equal(self.rr.sorted_values, [1, 2, 4])

    def test_same_kind(self):
        other = mm.SlidingTimeWindowReservoir(self.rr.window_size)
        nt.assert_true(self.rr.same_kind(other))

    def test_same_kind_with_different_class(self):
        other = mm.UniformReservoir(self.rr.window_size)
        nt.assert_false(self.rr.same_kind(other))

    def test_same_kind_with_different_parameters(self):
        other = mm.SlidingTimeWindowReservoir(10)
        nt.assert_false(self.rr.same_kind(other))


class TestExponentialDecayingReservoir(object):
    def setUp(self):
        self.patch = mock.patch('appmetrics.histogram.time.time', mock.Mock(return_value=0))
        self.time = self.patch.start()

        self.state = random.getstate()
        random.seed(42)

        self.size = 5
        self.rr = mm.ExponentialDecayingReservoir(self.size)

    def tearDown(self):
        self.patch.stop()
        random.setstate(self.state)

    @nt.raises(TypeError)
    def test_add_bad_type(self):
        self.rr.add(None)

    def _add_after(self, value, time):
        self.time.return_value += time
        self.rr.add(value)
        return self.rr.values

    def test_add_first(self):
        nt.assert_equal(self._add_after(1.5, 1), [1.5])
        nt.assert_equal(self._add_after(2.5, 1), [1.5, 2.5])
        nt.assert_equal(self._add_after(3.5, 1), [1.5, 3.5, 2.5])
        nt.assert_equal(self._add_after(4.5, 1), [1.5, 3.5, 4.5, 2.5])
        nt.assert_equal(self._add_after(5.5, 1), [5.5, 1.5, 3.5, 4.5, 2.5])

    def test_add_overflow(self):
        for i in range(1, 6):
            self._add_after(0.5+i, 1)

        nt.assert_equal(self._add_after(10, 1), [1.5, 10.0, 3.5, 4.5, 2.5])
        nt.assert_equal(self._add_after(20, 1), [1.5, 10.0, 3.5, 4.5, 2.5])
        nt.assert_equal(self._add_after(30, 1), [10.0, 3.5, 4.5, 30.0,  2.5])

    def test_rescaling(self):
        for i in range(1, 6):
            self._add_after(0.5+i, 1)

        # this should trigger a rescaling, so all the old values will have very small times
        # and all the new ones will be inserted
        self._add_after(10, 3600.0)
        for i in range(1, 5):
            self._add_after(10+i, 1)

        assert_items_equal(self.rr.values, [10, 11, 12, 13, 14])

    def test_long_delay(self):
        for i in range(1, 6):
            self._add_after(0.5+i, 1)

        # this emulates a new value after 15 hours: in that case the times are too small and collapse to zero
        nt.assert_equal(self._add_after(10, 3600.0*15), [2.5, 10.0])

    def test_sorted_values(self):
        self.rr._values = [(1, 10), (2, 2), (3, 1), (4, 4)]
        nt.assert_equal(self.rr.sorted_values, [1, 2, 4, 10])

    def test_same_kind(self):
        other = mm.ExponentialDecayingReservoir(self.rr.size)
        nt.assert_true(self.rr.same_kind(other))

    def test_same_kind_with_different_class(self):
        other = mm.UniformReservoir(self.rr.size)
        nt.assert_false(self.rr.same_kind(other))

    def test_same_kind_with_different_parameters(self):
        other = mm.ExponentialDecayingReservoir(10)
        nt.assert_false(self.rr.same_kind(other))

    def test_same_kind_with_different_parameters_2(self):
        other = mm.ExponentialDecayingReservoir(self.rr.size, 1)
        nt.assert_false(self.rr.same_kind(other))


class TestHistogram(object):
    def setUp(self):
        self.reservoir = mock.Mock()

        self.histogram = mm.Histogram(self.reservoir)

    def test_notify(self):
        result = self.histogram.notify(1.2)
        nt.assert_equal(
            self.reservoir.add.call_args_list,
            [mock.call(1.2)])
        nt.assert_equal(result, self.reservoir.add.return_value)

    def test_raw_data(self):
        result = self.histogram.raw_data()
        nt.assert_equal(result, self.reservoir.values)

    def test_get_values_zeros(self):
        self.reservoir.sorted_values = []

        expected = dict(
            kind="histogram",
            min=0,
            max=0,
            arithmetic_mean=0.0,
            geometric_mean=0.0,
            harmonic_mean=0.0,
            median=0.0,
            variance=0.0,
            standard_deviation=0.0,
            skewness=0.0,
            kurtosis=0.0,
            percentile=[(50, 0.0), (75, 0.0), (90, 0.0), (95, 0.0), (99, 0.0), (99.9, 0.0)],
            histogram=[(0, 0)],
            n=0
        )

        nt.assert_equal(self.histogram.get(), expected)

    def test_get_values(self):
        self.reservoir.sorted_values = [1.5, 2.5, 2.5, 2.75, 3.25, 3.26, 4.75]

        res = self.histogram.get()

        nt.assert_equal(res['kind'], "histogram")

        nt.assert_almost_equal(res['min'], 1.5)
        nt.assert_almost_equal(res['max'], 4.75)
        nt.assert_almost_equal(res['arithmetic_mean'], 2.93)
        nt.assert_almost_equal(res['geometric_mean'], 2.784379085700406)
        nt.assert_almost_equal(res['harmonic_mean'], 2.6362666258180956)
        nt.assert_almost_equal(res['median'], 2.75)
        nt.assert_almost_equal(res['variance'], 0.99513333)
        nt.assert_almost_equal(res['standard_deviation'], 0.9975636988851055)
        nt.assert_almost_equal(res['skewness'], 0.4329020512437358)
        nt.assert_almost_equal(res['kurtosis'], -0.8007344003569115)
        nt.assert_equal(res['percentile'], [(50, 2.75),
                                         (75, 3.25),
                                         (90, 3.26),
                                         (95, 4.75),
                                         (99, 4.75),
                                         (99.9, 4.75)])
        nt.assert_equal(res['histogram'], [(3.5, 6), (5.5, 1), (7.5, 0)])
        nt.assert_equal(res['n'], len(self.reservoir.sorted_values))




import csv
import os
import time
import tempfile
import shutil

from nose import tools as nt
import mock

from .. import reporter as mm, metrics, py3comp


class TestReporter(object):
    def setUp(self):
        self.old_registry = mm.REGISTRY
        mm.REGISTRY.clear()

    def tearDown(self):
        mm.REGISTRY.clear()
        mm.REGISTRY.update(self.old_registry)

    @nt.raises(TypeError)
    def test_register_invalid_schedule(self):
        mm.register(lambda x: None, None)

    @nt.raises(TypeError)
    def test_register_invalid_callback(self):
        mm.register(None, [1])

    @mock.patch('appmetrics.reporter.Timer')
    def test_register(self, timer):
        callback = mock.Mock()
        schedule = mock.MagicMock()
        tag = mock.Mock()

        mm.register(callback, schedule, tag)

        nt.assert_equal(len(mm.REGISTRY), 1)
        nt.assert_equal(list(mm.REGISTRY.values())[0], timer.return_value)
        nt.assert_equal(
            timer.call_args_list,
            [mock.call(schedule, callback, tag)])
        nt.assert_equal(
            timer.return_value.start.call_args_list,
            [mock.call()])

    @mock.patch('appmetrics.reporter.Timer')
    def test_register_with_default_tag(self, timer):
        callback = mock.Mock()
        schedule = mock.MagicMock()

        mm.register(callback, schedule)

        nt.assert_equal(len(mm.REGISTRY), 1)
        nt.assert_equal(list(mm.REGISTRY.values())[0], timer.return_value)
        nt.assert_equal(
            timer.call_args_list,
            [mock.call(schedule, callback, None)])
        nt.assert_equal(
            timer.return_value.start.call_args_list,
            [mock.call()])

    def test_get(self):
        mm.REGISTRY = {'m1': mock.Mock(), 'm2': mock.Mock()}
        nt.assert_equal(mm.get('m2'), mm.REGISTRY['m2'])

    def test_get_not_found(self):
        mm.REGISTRY = {'m1': mock.Mock(), 'm2': mock.Mock()}
        nt.assert_is_none(mm.get('m3'))

    def test_remove(self):
        m1 = mock.Mock()
        mm.REGISTRY = {'m1': m1, 'm2': mock.Mock()}

        nt.assert_is(mm.remove('m1'), m1)

        nt.assert_not_in('m1', mm.REGISTRY)
        nt.assert_equal(
            m1.cancel.call_args_list,
            [mock.call()])

    def test_remove_not_found(self):
        m1 = mock.Mock()
        m2 = mock.Mock()

        mm.REGISTRY = {'m1': m1, 'm2': m2}

        nt.assert_is_none(mm.remove('m3'))

        nt.assert_equal(mm.REGISTRY, dict(m1=m1, m2=m2))

    def test_fixed_interval_scheduler(self):
        now = time.time()
        sched = mm.fixed_interval_scheduler(10)

        nt.assert_almost_equal(next(sched), now+10, 0)
        nt.assert_almost_equal(next(sched), now+20, 0)
        nt.assert_almost_equal(next(sched), now+30, 0)

    def test_cleanup(self):
        m1 = mock.Mock()
        m2 = mock.Mock()

        mm.REGISTRY = {'m1': m1, 'm2': m2}

        mm.cleanup()

        nt.assert_equal(
            m1.cancel.call_args_list,
            [mock.call()])

        nt.assert_equal(
            m2.cancel.call_args_list,
            [mock.call()])


class TimeMachine(object):
    """
    Helper class to "emulate" the time
    """
    def __init__(self):
        self.current_time = 0
        self.sleeps = []

    def time(self):
        self.current_time += 1
        return self.current_time - 1

    def sleep(self, how_much):
        self.current_time += how_much
        self.sleeps.append(how_much)


class TestTimer(object):
    def setUp(self):
        self.time_patch = mock.patch('appmetrics.reporter.time', TimeMachine())
        self.time = self.time_patch.start()

        self.get_metrics_patch = mock.patch('appmetrics.reporter.get_metrics')
        self.get_metrics = self.get_metrics_patch.start()

        self.callback = mock.Mock()

        self.tag = "test"

    def tearDown(self):
        self.time_patch.stop()
        self.get_metrics_patch.stop()

    def test_finite_scheduler(self):
        def scheduler():
            yield 3
            yield 5
            yield 8

        tt = mm.Timer(scheduler(), self.callback, self.tag)
        tt.start()

        tt.join()

        nt.assert_false(tt.is_running)

        nt.assert_equal(
            self.callback.call_args_list,
            [
                mock.call(self.get_metrics.return_value),
                mock.call(self.get_metrics.return_value),
                mock.call(self.get_metrics.return_value),
            ]
        )

        nt.assert_equal(self.time.current_time, 10)
        nt.assert_equal(self.time.sleeps, [3, 1, 2])

    def test_infinite_scheduler(self):
        tt = mm.Timer(mm.fixed_interval_scheduler(10), self.callback, self.tag)

        # trick: when the time is >= 1000 call tt.cancel(), else it would run forever
        self.callback.side_effect = lambda x: None if self.time.current_time < 1000 else tt.cancel()

        tt.start()

        tt.join()

        nt.assert_false(tt.is_running)
        # one run every 10 seconds, for 1000 seconds total
        nt.assert_equal(self.callback.call_count, 1000/10)

        nt.assert_equal(self.time.sleeps, [11, 8] + [9]*98)

    def test_no_metrics(self):
        self.get_metrics.return_value = {}

        def scheduler():
            yield 3
            yield 5
            yield 8

        tt = mm.Timer(scheduler(), self.callback, self.tag)
        tt.start()

        tt.join()

        nt.assert_false(tt.is_running)

        nt.assert_equal(self.callback.call_count, 0)
        nt.assert_equal(self.time.sleeps, [3, 1, 2])


class TestGetMetrics(object):
    def setUp(self):
        self.original_registy = metrics.REGISTRY.copy()
        self.original_tags = metrics.TAGS.copy()

        self.m1, self.m2, self.m3 = (mock.Mock(), mock.Mock(), mock.Mock())

        metrics.REGISTRY = dict(m1=self.m1, m2=self.m2, m3=self.m3)
        metrics.TAGS = dict(tag={'m1', 'm3'})

    def tearDown(self):
        metrics.REGISTRY.clear()
        metrics.REGISTRY.update(self.original_registy)

        metrics.TAGS.clear()
        metrics.TAGS.update(self.original_tags)

    def test_with_none(self):
        res = mm.get_metrics(None)
        expected = dict(m1=self.m1.get(), m2=self.m2.get(), m3=self.m3.get())
        nt.assert_equal(res, expected)

    def test_with_tag(self):
        res = mm.get_metrics("tag")
        expected = dict(m1=self.m1.get(), m3=self.m3.get())
        nt.assert_equal(res, expected)

    def test_with_not_existent_tag(self):
        res = mm.get_metrics("xxx")
        nt.assert_equal(res, {})


class TestCSVReporter(object):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        metrics.new_histogram("h1")
        metrics.new_histogram("h2")
        metrics.new_meter("m1")
        metrics.new_meter("m2")
        metrics.new_gauge("g1")

        self.reporter = mm.CSVReporter(self.tmpdir)

        self.data = dict(
            h1=metrics.get("h1"),
            h2=metrics.get("h2"),
            m1=metrics.get("m1"),
            m2=metrics.get("m2"),
            g1=metrics.get("g1"),
        )

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        for name in ("h1", "h2", "m1", "m2", "g1"):
            metrics.delete_metric(name)

    def check_file(self, file_name, header, expected_data):
        with open(os.path.join(self.tmpdir, file_name), "r") as ff:
            reader = csv.reader(ff)
            data = list(reader)
            nt.assert_equal(data[0], list(header))
            nt.assert_equal(data[1:], expected_data)

    def histogram_data(self, time):
        return [time, "0", "0", "0"] + ["0.0"] * (len(mm.CSVReporter.histogram_header) - 4)

    def meter_data(self, time):
        return [time, "0", "0.0", "0.0", "0.0", "0.0", "0.0"]

    @mock.patch('appmetrics.reporter.time.time', mock.Mock(return_value=1234.5))
    def test_report_first(self):
        self.reporter(self.data)

        expected_files = ["h1_histogram.csv", "h2_histogram.csv", "m1_meter.csv", "m2_meter.csv"]

        py3comp.assert_items_equal(os.listdir(self.tmpdir), expected_files)

        hd = self.histogram_data("1234.5")
        md = self.meter_data("1234.5")

        self.check_file("h1_histogram.csv", mm.CSVReporter.histogram_header, [hd])
        self.check_file("h2_histogram.csv", mm.CSVReporter.histogram_header, [hd])

        self.check_file("m1_meter.csv", mm.CSVReporter.meter_header, [md])
        self.check_file("m2_meter.csv", mm.CSVReporter.meter_header, [md])

    @mock.patch('appmetrics.reporter.time.time')
    def test_report_multiple(self, time):
        times = [1150.2]*4 + [1140.2]*4 + [1130.2]*4
        time.side_effect = times.pop

        self.reporter(self.data)
        self.reporter(self.data)
        self.reporter(self.data)

        expected_files = ["h1_histogram.csv", "h2_histogram.csv", "m1_meter.csv", "m2_meter.csv"]

        py3comp.assert_items_equal(os.listdir(self.tmpdir), expected_files)

        hd = [self.histogram_data("1130.2"), self.histogram_data("1140.2"), self.histogram_data("1150.2")]
        md = [self.meter_data("1130.2"), self.meter_data("1140.2"), self.meter_data("1150.2")]

        self.check_file("h1_histogram.csv", mm.CSVReporter.histogram_header, hd)
        self.check_file("h2_histogram.csv", mm.CSVReporter.histogram_header, hd)

        self.check_file("m1_meter.csv", mm.CSVReporter.meter_header, md)
        self.check_file("m2_meter.csv", mm.CSVReporter.meter_header, md)



##  Module exceptions.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

"""
Exceptions module - export all the AppMetrics exceptions classes
"""


class AppMetricsError(Exception):
    """Base exception class"""
    pass


class StatisticsError(AppMetricsError, ValueError):
    """Statistic functions error"""
    pass


class DuplicateMetricError(AppMetricsError):
    """Raised if you are trying to create a metric with an already existing name"""
    pass


class InvalidMetricError(AppMetricsError, KeyError):
    """Raised if you are trying to use a metric that has not been registered"""
    pass




import mock
from nose.tools import assert_equal, assert_in, raises, assert_is, assert_is_instance, assert_false, assert_true

from .. import metrics as mm, exceptions, histogram, simple_metrics as simple, meter


class TestMetricsModule(object):
    def setUp(self):
        self.original_registy = mm.REGISTRY.copy()
        self.original_tags = mm.TAGS.copy()

        mm.REGISTRY.clear()
        mm.TAGS.clear()

    def tearDown(self):
        mm.REGISTRY.clear()
        mm.REGISTRY.update(self.original_registy)

        mm.TAGS.clear()
        mm.TAGS.update(self.original_tags)

    def test_new_metric(self):
        Cls = mock.Mock()

        args = [mock.Mock(), mock.Mock()]
        kwargs = dict(other=mock.Mock())

        res = mm.new_metric("test", Cls, *args, **kwargs)

        assert_in("test", mm.REGISTRY)

        item = mm.REGISTRY["test"]
        assert_equal(
            Cls.call_args_list,
            [mock.call(*args, **kwargs)]
        )
        assert_equal(item, Cls())
        assert_equal(item, res)

    @raises(exceptions.DuplicateMetricError)
    def test_new_metric_duplicated(self):
        Cls = mock.Mock()

        mm.new_metric("test", Cls)
        mm.new_metric("test", Cls)

    @raises(exceptions.InvalidMetricError)
    def test_metric_not_found(self):
        mm.metric("test")

    def test_metric(self):
        expected = mm.REGISTRY["test"] = mock.Mock()
        assert_equal(mm.metric("test"), expected)

    def test_metrics(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock())
        expected = ["test1", "test2"]
        assert_equal(mm.metrics(), expected)

    def test_get(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock())
        assert_equal(mm.get("test1"), mm.REGISTRY["test1"].get.return_value)

    @raises(exceptions.InvalidMetricError)
    def test_get_not_existing(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock())
        mm.get("test3")

    def test_notify(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock())
        mm.notify("test1", 123)
        assert_equal(
            mm.REGISTRY["test1"].notify.call_args_list,
            [mock.call(123)]
        )

    @raises(exceptions.InvalidMetricError)
    def test_notify_not_existing(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock())
        mm.notify("test3", 123)

    def test_delete_metric(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        mm.REGISTRY = dict(test1=m1, test2=m2)

        assert_equal(mm.delete_metric("test1"), m1)
        assert_equal(mm.REGISTRY, dict(test2=m2))

    def test_delete_metric_not_found(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        mm.REGISTRY = dict(test1=m1, test2=m2)

        assert_equal(mm.delete_metric("test3"), None)
        assert_equal(mm.REGISTRY, dict(test1=m1, test2=m2))

    def test_delete_metric_with_tags(self):
        mm.TAGS = {"test": {"test1", "test3"}}

        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()
        mm.REGISTRY = dict(test1=m1, test2=m2, test3=m3)

        assert_equal(mm.delete_metric("test1"), m1)
        assert_equal(mm.REGISTRY, dict(test2=m2, test3=m3))

        assert_equal(mm.TAGS["test"], {"test3"})

    def test_new_histogram_default(self):
        metric = mm.new_histogram("test")

        assert_is(metric, mm.metric("test"))

        assert_is_instance(metric, histogram.Histogram)
        assert_is_instance(metric.reservoir, histogram.UniformReservoir)
        assert_equal(metric.reservoir.size, histogram.DEFAULT_UNIFORM_RESERVOIR_SIZE)

    def test_new_histogram(self):
        metric = mm.new_histogram("test", histogram.UniformReservoir(10))

        assert_is(metric, mm.metric("test"))

        assert_is_instance(metric, histogram.Histogram)
        assert_is_instance(metric.reservoir, histogram.UniformReservoir)
        assert_equal(metric.reservoir.size, 10)

    def test_new_counter(self):
        metric = mm.new_counter("test")

        assert_is(metric, mm.metric("test"))

        assert_is_instance(metric, simple.Counter)

    def test_new_gauge(self):
        metric = mm.new_gauge("test")

        assert_is(metric, mm.metric("test"))

        assert_is_instance(metric, simple.Gauge)

    def test_new_meter(self):
        metric = mm.new_meter("test")

        assert_is(metric, mm.metric("test"))

        assert_is_instance(metric, meter.Meter)

    @raises(exceptions.InvalidMetricError)
    def test_new_reservoir_bad_type(self):
        mm.new_reservoir('xxx')

    @raises(TypeError)
    def test_new_reservoir_bad_args(self):
        mm.new_reservoir('uniform', xxx='yyy')

    def test_new_reservoir_with_defaults(self):
        reservoir = mm.new_reservoir()
        assert_is_instance(reservoir, histogram.UniformReservoir)
        assert_equal(reservoir.size, histogram.DEFAULT_UNIFORM_RESERVOIR_SIZE)

    def test_new_reservoir(self):
        reservoir = mm.new_reservoir('sliding_window', 5)
        assert_is_instance(reservoir, histogram.SlidingWindowReservoir)
        assert_equal(reservoir.size, 5)

    def test_new_histogram_with_implicit_reservoir(self):
        metric = mm.new_histogram_with_implicit_reservoir('test', 'sliding_window', 5)
        assert_is_instance(metric, histogram.Histogram)
        assert_is_instance(metric.reservoir, histogram.SlidingWindowReservoir)
        assert_equal(metric.reservoir.size, 5)

    @mock.patch('appmetrics.metrics.time')
    def test_with_histogram(self, time):
        # emulate the time spent in the function by patching time.time() and returning
        # two known values.
        times = [5, 3.4]
        time.time.side_effect = times.pop

        # decorated function
        @mm.with_histogram("test")
        def fun(v1, v2):
            """a docstring"""

            return v1+v2

        assert_equal(fun.__doc__, "a docstring")

        res = fun(1, 2)
        assert_equal(res, 3)

        assert_equal(mm.metric("test").raw_data(), [1.6])

    @mock.patch('appmetrics.metrics.time')
    def test_with_histogram_with_method(self, time):
        # emulate the time spent in the function by patching time.time() and returning
        # two known values.
        times = [5, 3.4]
        time.time.side_effect = times.pop

        # decorated method
        class MyClass(object):
            def __init__(self, v1):
                self.v1 = v1

            @mm.with_histogram("test")
            def method(self, v2):
                """a docstring"""

                return self.v1+v2

        assert_equal(MyClass.method.__doc__, "a docstring")

        obj = MyClass(1)

        assert_equal(obj.method.__doc__, "a docstring")

        res = obj.method(2)
        assert_equal(res, 3)

        assert_equal(mm.metric("test").raw_data(), [1.6])

    def test_with_histogram_multiple(self):
        @mm.with_histogram("test")
        def f1(v1, v2):
            """a docstring"""

            return v1+v2

        @mm.with_histogram("test")
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

        assert_equal(f1.__doc__, "a docstring")
        assert_equal(f2.__doc__, "another docstring")

        res = f1(1, 2)
        assert_equal(res, 3)

        res = f2(2, 3)
        assert_equal(res, 6)

        assert_equal(len(mm.metric("test").raw_data()), 2)

    @raises(exceptions.InvalidMetricError)
    def test_with_histogram_bad_reservoir_type(self):
        # decorated function
        @mm.with_histogram("test", "xxx")
        def fun(v1, v2):
            """a docstring"""

            return v1+v2

    @raises(exceptions.DuplicateMetricError)
    def test_with_histogram_multiple_and_arguments(self):
        @mm.with_histogram("test")
        def f1(v1, v2):
            """a docstring"""

            return v1+v2

        @mm.with_histogram("test", size=100)
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

    @raises(exceptions.DuplicateMetricError)
    def test_with_histogram_multiple_different_type(self):
        mm.new_gauge("test")

        @mm.with_histogram("test")
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

    def test_with_meter(self):

        @mm.with_meter("test")
        def fun(v):
            """a docstring"""

            return v*2

        assert_equal(fun.__doc__, "a docstring")

        res = [fun(i) for i in range(6)]
        assert_equal(res, [0, 2, 4, 6, 8, 10])

        assert_equal(mm.metric("test").raw_data(), 6)

    def test_with_meter_with_method(self):

        class MyClass(object):
            def __init__(self, v):
                self.v = v

            @mm.with_meter("test")
            def m1(self, v):
                """a docstring"""

                return v*self.v

            @mm.with_meter("test")
            def m2(self, v):
                """another docstring"""

                return v+self.v

        assert_equal(MyClass.m1.__doc__, "a docstring")
        assert_equal(MyClass.m2.__doc__, "another docstring")

        obj = MyClass(2)

        res = [obj.m1(i) for i in range(3)]
        assert_equal(res, [0, 2, 4])

        res = [obj.m2(i) for i in range(3)]
        assert_equal(res, [2, 3, 4])

        assert_equal(mm.metric("test").raw_data(), 6)

    def test_with_meter_multiple(self):
        @mm.with_meter("test")
        def f1(v1, v2):
            """a docstring"""

            return v1+v2

        @mm.with_meter("test")
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

        assert_equal(f1.__doc__, "a docstring")
        assert_equal(f2.__doc__, "another docstring")

        res = f1(1, 2)
        assert_equal(res, 3)

        res = f2(2, 3)
        assert_equal(res, 6)

        assert_equal(mm.metric("test").raw_data(), 2)

    @raises(exceptions.DuplicateMetricError)
    def test_with_meter_multiple_and_arguments(self):
        @mm.with_meter("test")
        def f1(v1, v2):
            """a docstring"""

            return v1+v2

        @mm.with_meter("test", tick_interval=100)
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

    @raises(exceptions.DuplicateMetricError)
    def test_with_meter_multiple_different_type(self):
        mm.new_gauge("test")

        @mm.with_meter("test")
        def f2(v1, v2):
            """another docstring"""

            return v1*v2

    @mock.patch('appmetrics.histogram.Histogram.notify')
    def test_timer(self, notify):
        with mm.timer("test"):
            pass

        assert_equal(notify.call_count, 1)

    @mock.patch('appmetrics.histogram.Histogram.notify')
    def test_timer_multiple(self, notify):
        with mm.timer("test"):
            pass

        with mm.timer("test"):
            pass

        assert_equal(notify.call_count, 2)

    @raises(exceptions.DuplicateMetricError)
    def test_timer_multiple_different_reservoir(self):
        with mm.timer("test", reservoir_type="sliding_window"):
            pass

        with mm.timer("test"):
            pass

    @raises(exceptions.DuplicateMetricError)
    def test_timer_multiple_different_type(self):
        mm.new_gauge("test")

        with mm.timer("test"):
            pass

    @raises(exceptions.InvalidMetricError)
    def test_tag_invalid_name(self):
        mm.tag("test", "test")

    def test_tag(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()

        mm.REGISTRY = {"test1": m1, "test2": m2, "test3": m3}
        mm.tag("test1", "1")
        mm.tag("test3", "1")
        mm.tag("test2", "2")

        assert_equal(mm.TAGS, {"1": {"test1", "test3"}, "2": {"test2"}})

    def test_tags(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_equal(mm.tags(), mm.TAGS)
        assert_false(mm.tags() is mm.TAGS)

    def test_untag_bad_tag(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_false(mm.untag("test1", "xxx"))

    def test_untag_bad_metric(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_false(mm.untag("xxx", "1"))

    def test_untag(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_true(mm.untag("test1", "1"))

        assert_equal(mm.TAGS, {"1": {"test3"}, "2": {"test2"}})

    def test_untag_last_group(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_true(mm.untag("test1", "1"))
        assert_true(mm.untag("test3", "1"))

        assert_equal(mm.TAGS, {"2": {"test2"}})

    def test_metrics_by_tag_invalid_tag(self):
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test2"}}

        assert_equal(mm.metrics_by_tag("test"), {})

    def test_metrics_by_tag(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()

        mm.REGISTRY = {"test1": m1, "test2": m2, "test3": m3}
        mm.TAGS = {"1": {"test1", "test3"}, "2": {"test3"}}

        assert_equal(mm.metrics_by_tag("1"), {"test1": m1.get(), "test3": m3.get()})

    def test_metrics_by_tag_deletion_while_looping(self):
        m1 = mock.Mock()
        m2 = mock.Mock()
        m3 = mock.Mock()

        m2.get.side_effect = exceptions.InvalidMetricError

        mm.REGISTRY = {"test1": m1, "test2": m2, "test3": m3}
        mm.TAGS = {"1": {"test1", "test2", "test3"}, "2": {"test2"}}


        assert_equal(mm.metrics_by_tag("1"), {"test1": m1.get(), "test3": m3.get()})

    def test_metrics_by_name_list(self):
        mm.REGISTRY = dict(test1=mock.Mock(), test2=mock.Mock(), test3=mock.Mock())
        out = mm.metrics_by_name_list(["test1", "test3"])
        expected = {'test1': mm.REGISTRY["test1"].get.return_value,
                    'test3': mm.REGISTRY["test3"].get.return_value}

        assert_equal(out, expected)




##  Module simple_metrics.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

"""
Implementation of simple metrics
"""

import threading


class Counter(object):
    """
    Counter metrics provide increment and decrement capabilities for a single integer value.
    """

    def __init__(self):
        self.value = 0

        self.lock = threading.Lock()

    def notify(self, value):
        """
        Increment or decrement the value, according to the given value's sign

        The value should be an integer, an attempt to cast it to integer will be made
        """
        value = int(value)

        with self.lock:
            self.value += value

    def get(self):
        """
        Return the counter's value
        """
        return dict(kind="counter", value=self.value)

    def raw_data(self):
        """
        Return the raw value
        """
        return self.value


class Gauge(object):
    """
    Gauges are point-in-time single value metrics.
    """

    def __init__(self):
        self.value = None

        self.lock = threading.Lock()

    def notify(self, value):
        """
        Set the current value for the gauge. The value may be any python value
        """

        with self.lock:
            self.value = value

    def get(self):
        """
        Return the gauge's current value
        """

        return dict(kind="gauge", value=self.value)

    def raw_data(self):
        return self.value



import math
from fractions import Fraction as F
from decimal import Decimal as D

from nose import tools as nt

from .. import statistics as mm
from ..exceptions import StatisticsError

"""adapted from original doctests"""

def test_isfinite():
    cases = [(0.0, True),
             (1.0, True),
             (5, True),
             (float("nan"), False),
             (float("inf"), False)]

    for v, expected in cases:
        check = lambda x: nt.assert_equal(mm.isfinite(x), expected)
        yield check, v

def test_coerce_types():
    class I1(int): pass
    class I2(I1): pass
    class I3(I1): pass
    class I4(I2): pass

    cases = [(int, float, float),
                (I1, I2, I2),
                (I2, I1, I2),
                (I1, float, float),
                (float, I1, float),
                (I2, I3, I3),
                (I4, I3, TypeError)]

    def check(x, y, expected):
        if issubclass(expected, Exception):
            nt.assert_raises(expected, mm.coerce_types, x, y)
        else:
            nt.assert_is(mm.coerce_types(x, y), expected)

    for x, y, res in cases:
        yield check, x, y, res

def test_sum_plain():
    value = mm.sum([3, 2.25, 4.5, -0.5, 1.0], 0.75)
    nt.assert_equal(value, 11.0)

def test_sum_extreme():
    value = mm.sum([1e50, 1, -1e50] * 1000)  # Built-in sum returns zero.
    nt.assert_equal(value, 1000.0)

def test_sum_with_fractions():
    value = mm.sum([F(2, 3), F(7, 5), F(1, 4), F(5, 6)])
    nt.assert_equal(value, F(63, 20))

def test_sum_with_decimals():
    data = [D("0.1375"), D("0.2108"), D("0.3061"), D("0.0419")]
    nt.assert_equal(mm.sum(data), D('0.6963'))

def test_sum_with_nan():
    nt.assert_true(math.isnan(mm.sum([1, float("nan")])))

def test_exact_ratio():
    nt.assert_equal(mm.exact_ratio(0.25), (1, 4))

@nt.raises(TypeError)
def test_exact_ratio_None():
    mm.exact_ratio(None)

def test_exact_ratio_nan():
    n, d = mm.exact_ratio(float("nan"))
    nt.assert_true(math.isnan(n))
    nt.assert_is_none(d)

def test_decimal_to_ratio():
    nt.assert_equal(mm.decimal_to_ratio(D("2.6")), (26, 10))

@nt.raises(ValueError)
def test_decimal_to_ratio_nan():
    mm.decimal_to_ratio(D("nan"))

@nt.raises(TypeError)
def test_counts_not_iterable():
    mm.counts(None)

def test_counts():
    nt.assert_equal(mm.counts([1, 2, 2, 3, 2, 4, 2]), [(2, 4)])

def test_mean():
    nt.assert_equal(mm.mean([1, 2, 3, 4, 4]), 2.8)

def test_mean_fractions():
    value = mm.mean([F(3, 7), F(1, 21), F(5, 3), F(1, 3)])
    nt.assert_equal(value, F(13, 21))

def test_mean_decimals():
    value = mm.mean([D("0.5"), D("0.75"), D("0.625"), D("0.375")])
    nt.assert_equal(value, D("0.5625"))

def test_mean_iter():
    value = mm.mean(iter([1, 2, 3]))
    nt.assert_equal(value, 2)

    value = mm.mean(range(3))
    nt.assert_equal(value, 1)

def test_median():
    nt.assert_equal(mm.median([1, 3, 5]), 3)
    nt.assert_equal(mm.median([1, 3, 5, 7]), 4)

def test_median_low():
    nt.assert_equal(mm.median_low([1, 3, 5]), 3)
    nt.assert_equal(mm.median_low([1, 3, 5, 7]), 3)

@nt.raises(StatisticsError)
def test_median_low_empty():
    mm.median_low([])

def test_median_high():
    nt.assert_equal(mm.median_high([1, 3, 5]), 3)
    nt.assert_equal(mm.median_high([1, 3, 5, 7]), 5)

@nt.raises(StatisticsError)
def test_median_high_empty():
    mm.median_high([])

def test_mode():
    nt.assert_equal(mm.mode([1, 1, 2, 3, 3, 3, 3, 4]), 3)

def test_mode_strings():
    nt.assert_equal(mm.mode(["red", "blue", "blue", "red", "green", "red", "red"]), "red")

@nt.raises(StatisticsError)
def test_mode_not_unique():
    mm.mode(["red", "blue", "blue", "red"])

@nt.raises(StatisticsError)
def test_mode_empty():
    mm.mode([])

def test_variance():
    data = [2.75, 1.75, 1.25, 0.25, 0.5, 1.25, 3.5]
    nt.assert_equal(mm.variance(data), 1.3720238095238095)

def test_variance_with_mean():
    data = [2.75, 1.75, 1.25, 0.25, 0.5, 1.25, 3.5]
    mean = mm.mean(data)
    nt.assert_equal(mm.variance(data, mean), 1.3720238095238095)

def test_variance_decimals():
    value = mm.variance([D("27.5"), D("30.25"), D("30.25"), D("34.5"), D("41.75")])
    nt.assert_equal(value, D('31.01875'))

def test_variance_fractions():
    value = mm.variance([F(1, 6), F(1, 2), F(5, 3)])
    nt.assert_equal(value, F(67, 108))

def test_variance_iterator():
    nt.assert_equal(mm.variance(iter([1, 2, 5])), 4.333333333333334)

def test_pvariance():
    data = [0.0, 0.25, 0.25, 1.25, 1.5, 1.75, 2.75, 3.25]
    nt.assert_equal(mm.pvariance(data), 1.25)

def test_pvariance_with_mean():
    data = [0.0, 0.25, 0.25, 1.25, 1.5, 1.75, 2.75, 3.25]
    mean = mm.mean(data)
    nt.assert_equal(mm.pvariance(data, mean), 1.25)

def test_pvariance_decimals():
    value = mm.pvariance([D("27.5"), D("30.25"), D("30.25"), D("34.5"), D("41.75")])
    nt.assert_equal(value, D('24.815'))

def test_pvariance_fractions():
    value = mm.pvariance([F(1, 4), F(5, 4), F(1, 2)])
    nt.assert_equal(value, F(13, 72))

def test_pvariance_iterator():
    nt.assert_equal(mm.pvariance(iter([1, 2, 3])), 0.6666666666666666)

@nt.raises(StatisticsError)
def test_pvariance_empty():
    mm.pvariance([])

def test_stdev():
    value = mm.stdev([1.5, 2.5, 2.5, 2.75, 3.25, 4.75])
    nt.assert_equal(value, 1.0810874155219827)

def test_pstddev():
    value = mm.pstdev([1.5, 2.5, 2.5, 2.75, 3.25, 4.75])
    nt.assert_equal(value, 0.986893273527251)

def test_geometric_mean():
    cases = [
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 2.7121486566834387),
        ([1.5, 2.5, 2.5, 2.75, -3.25, 4.75], 2.2284330795786698),
        ([1.5, 2.5, 2.5, 2.75, 0, 4.75], 2.63258262293452),]

    for data, expected in cases:
        fun = lambda data, expected: nt.assert_almost_equal(mm.geometric_mean(data), expected)
        yield fun, data, expected

@nt.raises(StatisticsError)
def test_geometric_mean_empty():
    mm.geometric_mean([])

def test_harmonic_mean():
    cases = [
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 2.5547986710408086),
        ([1.5, 2.5, 2.5, 2.75, -3.25, 4.75], 3.4619305150494095),
        ([1.5, 2.5, 2.5, 2.75, 0, 4.75], 2.9399812441387936),]

    for data, expected in cases:
        fun = lambda data, expected: nt.assert_almost_equal(mm.harmonic_mean(data), expected)
        yield fun, data, expected

@nt.raises(StatisticsError)
def test_harmonic_mean_empty():
    mm.harmonic_mean([])

def test_skewness():
    cases = [
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 0.5193819834930741),
        ([1.5, 2.5, 2.5, 2.75, -3.25, 4.75], -0.8623011014727693),
        ([1.5, 2.5, 2.5, 2.75, 0, 4.75], 0.03989169123327498),
        ([1.0, 1.0, 1.0], 0.0),
    ]

    for data, expected in cases:
        fun = lambda data, expected: nt.assert_almost_equal(mm.skewness(data), expected)
        yield fun, data, expected

@nt.raises(StatisticsError)
def test_skewness_empty():
    mm.skewness([])

def test_kurtosis():
    cases = [
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], -1.0485692089183762),
        ([1.5, 2.5, 2.5, 2.75, -3.25, 4.75], -0.6949222688831207),
        ([1.5, 2.5, 2.5, 2.75, 0, 4.75], -1.2034104454720884),
        ([1.0, 1.0, 1.0], 0.0),
    ]

    for data, expected in cases:
        fun = lambda data, expected: nt.assert_almost_equal(mm.kurtosis(data), expected)
        yield fun, data, expected

@nt.raises(StatisticsError)
def test_kurtosis_empty():
    mm.kurtosis([])

def test_percentile():
    cases = [
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 50, 2.5),
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 99, 4.75),
        ([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 10, 1.5),
    ]

    for data, confidence, expected in cases:
        fun = lambda data, confidence, expected: nt.assert_almost_equal(mm.percentile(data, confidence), expected)
        yield fun, data, confidence, expected


def test_percentile_few_points():
    with nt.assert_raises_regexp(StatisticsError, "few data points \(6\) for 1th"):
        mm.percentile([1.5, 2.5, 2.5, 2.75, 3.25, 4.75], 1)

def test_percentile_empty():
    with nt.assert_raises_regexp(StatisticsError, "few data points \(0\) for 90th"):
        mm.percentile([], 90)

def test_get_histogram():
    value = mm.get_histogram([1.5, 2.5, 2.5, 2.75, 3.25, 4.75, 5.0])
    nt.assert_equal(value, [(3.5, 5), (5.5, 2), (7.5, 0)])

    value = mm.get_histogram([1.0, 1.0, 1.0])
    nt.assert_equal(value, [(2.0, 3)])

def test_get_histogram_few_points():
    with nt.assert_raises_regexp(StatisticsError, "few data points \(1\) for get_histogram"):
        mm.get_histogram([1.5])

def test_get_histogram_bins():
    nt.assert_equal(mm.get_histogram_bins(1, 3, 0.5, 5), [2, 3, 4])

    nt.assert_equal(mm.get_histogram_bins(1, 2, -1/3.5, 1.0), [1])



##  Module metrics.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

"""
Main interface module
"""

from contextlib import contextmanager
import functools
import threading
import time

from .exceptions import DuplicateMetricError, InvalidMetricError
from . import histogram, simple_metrics, meter, py3comp


REGISTRY = {}
TAGS = {}
LOCK = threading.Lock()


def new_metric(name, class_, *args, **kwargs):
    """Create a new metric of the given class.

    Raise DuplicateMetricError if the given name has been already registered before

    Internal function - use "new_<type> instead"
    """

    with LOCK:
        try:
            item = REGISTRY[name]
        except KeyError:
            item = REGISTRY[name] = class_(*args, **kwargs)
            return item

    raise DuplicateMetricError("Metric {} already exists of type {}".format(name, type(item).__name__))


def delete_metric(name):
    """Remove the named metric"""

    with LOCK:
        old_metric = REGISTRY.pop(name, None)

        # look for the metric name in the tags and remove it
        for _, tags in py3comp.iteritems(TAGS):
            if name in tags:
                tags.remove(name)

    return old_metric


def metric(name):
    """
    Return the metric with the given name, if any

    Raise InvalidMetricError if the given name has not been registered
    """

    try:
        return REGISTRY[name]
    except KeyError as e:
        raise InvalidMetricError("Metric {} not found!".format(e))


def metrics():
    """
    Return the list of the returned metrics' names
    """

    return sorted(REGISTRY.keys())


def get(name):
    """
    Call "get" on the metric with the given name
    Raise InvalidMetricError if the given name has not been registered
    """

    return metric(name).get()


def notify(name, value):
    """
    Call "notify" on the metric with the given name
    Raise InvalidMetricError if the given name has not been registered
    """

    return metric(name).notify(value)


def new_histogram(name, reservoir=None):
    """
    Build a new histogram metric with a given reservoir object
    If the reservoir is not provided, a uniform reservoir with the default size is used
    """

    if reservoir is None:
        reservoir = histogram.UniformReservoir(histogram.DEFAULT_UNIFORM_RESERVOIR_SIZE)

    return new_metric(name, histogram.Histogram, reservoir)


def new_counter(name):
    """
    Build a new "counter" metric
    """

    return new_metric(name, simple_metrics.Counter)


def new_gauge(name):
    """
    Build a new "gauge" metric
    """

    return new_metric(name, simple_metrics.Gauge)


def new_meter(name, tick_interval=5):
    """
    Build a new "meter" metric
    """

    return new_metric(name, meter.Meter, tick_interval)


def new_histogram_with_implicit_reservoir(name, reservoir_type='uniform', *reservoir_args, **reservoir_kwargs):
    """
    Build a new histogram metric and a reservoir from the given parameters
    """

    reservoir = new_reservoir(reservoir_type, *reservoir_args, **reservoir_kwargs)
    return new_histogram(name, reservoir)


def new_reservoir(reservoir_type='uniform', *reservoir_args, **reservoir_kwargs):
    """
    Build a new reservoir
    """

    try:
        reservoir_cls = RESERVOIR_TYPES[reservoir_type]
    except KeyError:
        raise InvalidMetricError("Unknown reservoir type: {}".format(reservoir_type))

    return reservoir_cls(*reservoir_args, **reservoir_kwargs)


def get_or_create_histogram(name, reservoir_type, *reservoir_args, **reservoir_kwargs):
    """
    Will return a histogram matching the given parameters or raise
    DuplicateMetricError if it can't be created due to a name collision
    with another histogram with different parameters.
    """
    reservoir = new_reservoir(reservoir_type, *reservoir_args, **reservoir_kwargs)

    try:
        hmetric = new_histogram(name, reservoir)
    except DuplicateMetricError:
        hmetric = metric(name)
        if not isinstance(hmetric, histogram.Histogram):
            raise DuplicateMetricError(
                "Metric {!r} already exists of type {!r}".format(name, type(hmetric).__name__))

        if not hmetric.reservoir.same_kind(reservoir):
            raise DuplicateMetricError(
                "Metric {!r} already exists with a different reservoir: {}".format(name, hmetric.reservoir))

    return hmetric


def with_histogram(name, reservoir_type="uniform", *reservoir_args, **reservoir_kwargs):
    """
    Time-measuring decorator: the time spent in the wrapped function is measured
    and added to the named metric.
    metric_args and metric_kwargs are passed to new_histogram()
    """

    hmetric = get_or_create_histogram(name, reservoir_type, *reservoir_args, **reservoir_kwargs)

    def wrapper(f):

        @functools.wraps(f)
        def fun(*args, **kwargs):
            t1 = time.time()
            res = f(*args, **kwargs)
            t2 = time.time()

            hmetric.notify(t2-t1)
            return res

        return fun

    return wrapper


def with_meter(name, tick_interval=meter.DEFAULT_TICK_INTERVAL):
    """
    Call-counting decorator: each time the wrapped function is called
    the named meter is incremented by one.
    metric_args and metric_kwargs are passed to new_meter()
    """

    try:
        mmetric = new_meter(name, tick_interval)
    except DuplicateMetricError as e:
        mmetric = metric(name)

        if not isinstance(mmetric, meter.Meter):
            raise DuplicateMetricError("Metric {!r} already exists of type {}".format(name, type(mmetric).__name__))

        if tick_interval != mmetric.tick_interval:
            raise DuplicateMetricError("Metric {!r} already exists: {}".format(name, mmetric))

    def wrapper(f):

        @functools.wraps(f)
        def fun(*args, **kwargs):
            res = f(*args, **kwargs)

            mmetric.notify(1)
            return res

        return fun

    return wrapper


@contextmanager
def timer(name, reservoir_type="uniform", *reservoir_args, **reservoir_kwargs):
    """
    Time-measuring context manager: the time spent in the wrapped block
    if measured and added to the named metric.
    """

    hmetric = get_or_create_histogram(name, reservoir_type, *reservoir_args, **reservoir_kwargs)

    t1 = time.time()
    yield
    t2 = time.time()
    hmetric.notify(t2 - t1)


def tag(name, tag_name):
    """
    Tag the named metric with the given tag.
    """

    with LOCK:
        # just to check if <name> exists
        metric(name)

        TAGS.setdefault(tag_name, set()).add(name)


def tags():
    """
    Return the currently defined tags.
    """

    # protect global value against accidental modifications
    return TAGS.copy()


def metrics_by_tag(tag_name):
    """
    Return a dictionary with {metric name: metric values} for all the metrics with the given tag.
    Return an empty dictionary if the given tag does not exist.
    """

    try:
        names = TAGS[tag_name]
    except KeyError:
        return {}

    return metrics_by_name_list(names)


def untag(name, tag_name):
    """
    Remove the given tag from the given metric.
    Return True if the metric was tagged, False otherwise
    """

    with LOCK:
        by_tag = TAGS.get(tag_name, None)
        if not by_tag:
            return False
        try:
            by_tag.remove(name)

            # remove the tag if no associations left
            if not by_tag:
                TAGS.pop(tag_name)

            return True
        except KeyError:
            return False


def metrics_by_name_list(names):
    """
    Return a dictionary with {metric name: metric value} for all the metrics with the given names.
    """
    results = {}

    for name in names:
        # no lock - a metric could have been removed in the meanwhile
        try:
            results[name] = get(name)
        except InvalidMetricError:
            continue

    return results


RESERVOIR_TYPES = {
    'uniform': histogram.UniformReservoir,
    'sliding_window': histogram.SlidingWindowReservoir,
    'sliding_time_window': histogram.SlidingTimeWindowReservoir,
    'exp_decaying': histogram.ExponentialDecayingReservoir,
}


METRIC_TYPES = {
    'histogram': new_histogram_with_implicit_reservoir,
    'gauge': new_gauge,
    'counter': new_counter,
    'meter': new_meter,
}


from setuptools import setup, find_packages

setup(
    name = "AppMetrics",
    version = "0.5.0",

    packages = find_packages(),

    scripts = [],

    install_requires = [],

    package_data = {
    },

    # metadata for upload to PyPI
    author = "Antonio Valente",
    author_email = "y3sman@gmail.com",
    description = "Application metrics collector",
    license = "Apache 2.0",
    keywords = ["metrics", "folsom", "histogram", "metric", "monitor"],
    url = "https://github.com/avalente/appmetrics",
    platforms = 'Platform Independent',

    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: System :: Monitoring",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
    ]
)



##  Module statistics.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.


"""
Statistics module.

The basic functions are stolen from python 3.4 stdlib

"""

from __future__ import division

import collections
import math
import operator
import functools

from fractions import Fraction
from decimal import Decimal

from .exceptions import StatisticsError
from .py3comp import xrange, iteritems


def isfinite(n):
    """Return True if x is neither an infinity nor a NaN, and False otherwise.
    (Note that 0.0 is considered finite.)

    Backported from python 3
    """

    return not (math.isinf(n) or math.isnan(n))


def sum(data, start=0):
    """sum(data [, start]) -> value

    Return a high-precision sum of the given numeric data. If optional
    argument ``start`` is given, it is added to the total. If ``data`` is
    empty, ``start`` (defaulting to 0) is returned.
    """
    n, d = exact_ratio(start)
    T = type(start)
    partials = {d: n}  # map {denominator: sum of numerators}
    # Micro-optimizations.
    coerce_types_ = coerce_types
    exact_ratio_ = exact_ratio
    partials_get = partials.get
    # Add numerators for each denominator, and track the "current" type.
    for x in data:
        T = coerce_types_(T, type(x))
        n, d = exact_ratio_(x)
        partials[d] = partials_get(d, 0) + n
    if None in partials:
        assert issubclass(T, (float, Decimal))
        assert not isfinite(partials[None])
        return T(partials[None])
    total = Fraction()
    for d, n in sorted(partials.items()):
        total += Fraction(n, d)
    if issubclass(T, int):
        assert total.denominator == 1
        return T(total.numerator)
    if issubclass(T, Decimal):
        return T(total.numerator) / total.denominator
    return T(total)


def exact_ratio(x):
    """Convert Real number x exactly to (numerator, denominator) pair.

    x is expected to be an int, Fraction, Decimal or float.
    """
    try:
        try:
            # int, Fraction
            return x.numerator, x.denominator
        except AttributeError:
            # float
            try:
                return x.as_integer_ratio()
            except AttributeError:
                # Decimal
                try:
                    return decimal_to_ratio(x)
                except AttributeError:
                    msg = "can't convert type '{}' to numerator/denominator"
                    raise TypeError(msg.format(type(x).__name__))
    except (OverflowError, ValueError):
        # INF or NAN
        return (x, None)


# FIXME This is faster than Fraction.from_decimal, but still too slow.
def decimal_to_ratio(d):
    """Convert Decimal d to exact integer ratio (numerator, denominator).
    """
    sign, digits, exp = d.as_tuple()
    if exp in ('F', 'n', 'N'):  # INF, NAN, sNAN
        assert not d.is_finite()
        raise ValueError
    num = 0
    for digit in digits:
        num = num * 10 + digit
    if sign:
        num = -num
    den = 10 ** -exp
    return (num, den)


def coerce_types(T1, T2):
    """Coerce types T1 and T2 to a common type.

    Coercion is performed according to this table, where "N/A" means
    that a TypeError exception is raised.

    +----------+-----------+-----------+-----------+----------+
    |          | int       | Fraction  | Decimal   | float    |
    +----------+-----------+-----------+-----------+----------+
    | int      | int       | Fraction  | Decimal   | float    |
    | Fraction | Fraction  | Fraction  | N/A       | float    |
    | Decimal  | Decimal   | N/A       | Decimal   | float    |
    | float    | float     | float     | float     | float    |
    +----------+-----------+-----------+-----------+----------+

    Subclasses trump their parent class; two subclasses of the same
    base class will be coerced to the second of the two.

    """
    # Get the common/fast cases out of the way first.
    if T1 is T2: return T1
    if T1 is int: return T2
    if T2 is int: return T1
    # Subclasses trump their parent class.
    if issubclass(T2, T1): return T2
    if issubclass(T1, T2): return T1
    # Floats trump everything else.
    if issubclass(T2, float): return T2
    if issubclass(T1, float): return T1
    # Subclasses of the same base class give priority to the second.
    if T1.__base__ is T2.__base__: return T2
    # Otherwise, just give up.
    raise TypeError('cannot coerce types %r and %r' % (T1, T2))


def counts(data):
    """
    Generate a table of sorted (value, frequency) pairs.
    """
    if data is None:
        raise TypeError('None is not iterable')
    table = collections.Counter(data).most_common()
    if not table:
        return table
        # Extract the values with the highest frequency.
    maxfreq = table[0][1]
    for i in range(1, len(table)):
        if table[i][1] != maxfreq:
            table = table[:i]
            break
    return table


# === Measures of central tendency (averages) ===

def mean(data):
    """Return the sample arithmetic mean of data.

    If ``data`` is empty, StatisticsError will be raised.
    """
    if iter(data) is data:
        data = list(data)
    n = len(data)
    if n < 1:
        raise StatisticsError('mean requires at least one data point')
    return sum(data) / n


# FIXME: investigate ways to calculate medians without sorting? Quickselect?
def median(data):
    """Return the median (middle value) of numeric data.

    When the number of data points is odd, return the middle data point.
    When the number of data points is even, the median is interpolated by
    taking the average of the two middle values:

    """
    data = sorted(data)
    n = len(data)
    if n == 0:
        raise StatisticsError("no median for empty data")
    if n % 2 == 1:
        return data[n // 2]
    else:
        i = n // 2
        return (data[i - 1] + data[i]) / 2


def median_low(data):
    """Return the low median of numeric data.

    When the number of data points is odd, the middle value is returned.
    When it is even, the smaller of the two middle values is returned.
    """
    data = sorted(data)
    n = len(data)
    if n == 0:
        raise StatisticsError("no median for empty data")
    if n % 2 == 1:
        return data[n // 2]
    else:
        return data[n // 2 - 1]


def median_high(data):
    """Return the high median of data.

    When the number of data points is odd, the middle value is returned.
    When it is even, the larger of the two middle values is returned.

    """
    data = sorted(data)
    n = len(data)
    if n == 0:
        raise StatisticsError("no median for empty data")
    return data[n // 2]


def mode(data):
    """Return the most common data point from discrete or nominal data.

    ``mode`` assumes discrete data, and returns a single value. This is the
    standard treatment of the mode as commonly taught in schools:

    If there is not exactly one most common value, ``mode`` will raise
    StatisticsError.
    """
    # Generate a table of sorted (value, frequency) pairs.
    table = counts(data)
    if len(table) == 1:
        return table[0][0]
    elif table:
        raise StatisticsError(
            'no unique mode; found %d equally common values' % len(table)
        )
    else:
        raise StatisticsError('no mode for empty data')


# === Measures of spread ===

# See http://mathworld.wolfram.com/Variance.html
#     http://mathworld.wolfram.com/SampleVariance.html
#     http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
#
# Under no circumstances use the so-called "computational formula for
# variance", as that is only suitable for hand calculations with a small
# amount of low-precision data. It has terrible numeric properties.
#
# See a comparison of three computational methods here:
# http://www.johndcook.com/blog/2008/09/26/comparing-three-methods-of-computing-standard-deviation/

def _ss(data, c=None):
    """Return sum of square deviations of sequence data.

    If ``c`` is None, the mean is calculated in one pass, and the deviations
    from the mean are calculated in a second pass. Otherwise, deviations are
    calculated from ``c`` as given. Use the second case with care, as it can
    lead to garbage results.
    """
    if c is None:
        c = mean(data)
    ss = sum((x - c) ** 2 for x in data)
    # The following sum should mathematically equal zero, but due to rounding
    # error may not.
    ss -= sum((x - c) for x in data) ** 2 / len(data)
    assert not ss < 0, 'negative sum of square deviations: %f' % ss
    return ss


def variance(data, xbar=None):
    """Return the sample variance of data.

    data should be an iterable of Real-valued numbers, with at least two
    values. The optional argument xbar, if given, should be the mean of
    the data. If it is missing or None, the mean is automatically calculated.

    Use this function when your data is a sample from a population. To
    calculate the variance from the entire population, see ``pvariance``.

    If you have already calculated the mean of your data, you can pass it as
    the optional second argument ``xbar`` to avoid recalculating it:

    This function does not check that ``xbar`` is actually the mean of
    ``data``. Giving arbitrary values for ``xbar`` may lead to invalid or
    impossible results.

    Decimals and Fractions are supported

    """
    if iter(data) is data:
        data = list(data)
    n = len(data)
    if n < 2:
        raise StatisticsError('variance requires at least two data points')
    ss = _ss(data, xbar)
    return ss / (n - 1)


def pvariance(data, mu=None):
    """Return the population variance of ``data``.

    data should be an iterable of Real-valued numbers, with at least one
    value. The optional argument mu, if given, should be the mean of
    the data. If it is missing or None, the mean is automatically calculated.

    Use this function to calculate the variance from the entire population.
    To estimate the variance from a sample, the ``variance`` function is
    usually a better choice.

    If you have already calculated the mean of the data, you can pass it as
    the optional second argument to avoid recalculating it:


    This function does not check that ``mu`` is actually the mean of ``data``.
    Giving arbitrary values for ``mu`` may lead to invalid or impossible
    results.

    Decimals and Fractions are supported:


    """
    if iter(data) is data:
        data = list(data)
    n = len(data)
    if n < 1:
        raise StatisticsError('pvariance requires at least one data point')
    ss = _ss(data, mu)
    return ss / n


def stdev(data, xbar=None):
    """Return the square root of the sample variance.

    See ``variance`` for arguments and other details.

    """
    var = variance(data, xbar)
    try:
        return var.sqrt()
    except AttributeError:
        return math.sqrt(var)


def pstdev(data, mu=None):
    """Return the square root of the population variance.

    See ``pvariance`` for arguments and other details.

    """
    var = pvariance(data, mu)
    try:
        return var.sqrt()
    except AttributeError:
        return math.sqrt(var)


def geometric_mean(data):
    """Return the geometric mean of data
    """

    if not data:
        raise StatisticsError('geometric_mean requires at least one data point')

    # in order to support negative or null values
    data = [x if x > 0 else math.e if x == 0 else 1.0 for x in data]

    return math.pow(math.fabs(functools.reduce(operator.mul, data)), 1.0 / len(data))


def harmonic_mean(data):
    """Return the harmonic mean of data
    """

    if not data:
        raise StatisticsError('harmonic_mean requires at least one data point')

    return len(data) / sum(map(lambda x: 1.0 / x if x else 0.0, data))


def skewness(data):
    """Return the skewness of the data's distribution

    """

    if not data:
        raise StatisticsError('skewness requires at least one data point')

    size = len(data)
    sd = stdev(data) ** 3

    if not sd:
        return 0.0

    mn = mean(data)
    return sum(map(lambda x: ((x - mn) ** 3 / sd), data)) / size


def kurtosis(data):
    """Return the kurtosis of the data's distribution

    """

    if not data:
        raise StatisticsError('kurtosis requires at least one data point')

    size = len(data)
    sd = stdev(data) ** 4

    if not sd:
        return 0.0

    mn = mean(data)
    return sum(map(lambda x: ((x - mn) ** 4 / sd), data)) / size - 3


def percentile(data, n):
    """Return the n-th percentile of the given data

    Assume that the data are already sorted

    """

    size = len(data)
    idx = (n / 100.0) * size - 0.5

    if idx < 0 or idx > size:
        raise StatisticsError("Too few data points ({}) for {}th percentile".format(size, n))

    return data[int(idx)]


def get_histogram(data):
    """Return the histogram relative to the given data

    Assume that the data are already sorted

    """

    count = len(data)

    if count < 2:
        raise StatisticsError('Too few data points ({}) for get_histogram'.format(count))

    min_ = data[0]
    max_ = data[-1]
    std = stdev(data)

    bins = get_histogram_bins(min_, max_, std, count)

    res = {x: 0 for x in bins}

    for value in data:
        for bin_ in bins:
            if value <= bin_:
                res[bin_] += 1
                break

    return sorted(iteritems(res))


def get_histogram_bins(min_, max_, std, count):
    """
    Return optimal bins given the input parameters

    """

    width = _get_bin_width(std, count)
    count = int(round((max_ - min_) / width) + 1)

    if count:
        bins = [i * width + min_ for i in xrange(1, count + 1)]
    else:
        bins = [min_]

    return bins


def _get_bin_width(stdev, count):
    """Return the histogram's optimal bin width based on Sturges

    http://www.jstor.org/pss/2965501
    """

    w = int(round((3.5 * stdev) / (count ** (1.0 / 3))))
    if w:
        return w
    else:
        return 1




import logging

from nose.tools import assert_equal, assert_almost_equal, raises
import mock

from .. import meter as mm


log = logging.getLogger(__name__)


class TestEWMA(object):
    def test_compute_alpha(self):
        assert_almost_equal(mm.EWMA.compute_alpha(1, 5), 0.07995558537067671)
        assert_almost_equal(mm.EWMA.compute_alpha(5, 5), 0.01652854617838251)
        assert_almost_equal(mm.EWMA.compute_alpha(60 * 24, 5), 5.7868695912732804e-5)

    def test_update(self):
        obj = mm.EWMA(1, 5)
        obj.update(1)
        obj.update(5)

        assert_equal(obj.value, 6)

    @raises(ValueError)
    def test_update_bad_value(self):
        obj = mm.EWMA(1, 5)
        obj.update("xxx")

    def test_tick_first_time(self):
        obj = mm.EWMA(1, 5)
        obj.value = 5

        obj.tick()

        assert_equal(obj.value, 0)
        assert_almost_equal(obj.rate, 1.0)

        obj.tick()

        assert_equal(obj.value, 0)
        assert_almost_equal(obj.rate, 0.9200444146293233)

        obj.tick()

        assert_equal(obj.value, 0)
        assert_almost_equal(obj.rate, 0.8464817248906141)


class TestMeter(object):
    def setUp(self):
        self.meter = mm.Meter()
        self.started_on = self.meter.started_on

    def test_notify(self):
        self.meter.tick = mock.Mock()

        self.meter.notify(1)
        assert_equal(self.meter.count, 1)
        assert_equal(self.meter.m1.value, 1)
        assert_equal(self.meter.m5.value, 1)
        assert_equal(self.meter.m15.value, 1)
        assert_equal(self.meter.day.value, 1)

        assert_equal(self.meter.tick.call_args_list, [[]])

    def test_tick_all(self):
        self.meter.m1 = mock.Mock()
        self.meter.m5 = mock.Mock()
        self.meter.m15 = mock.Mock()
        self.meter.day = mock.Mock()

        self.meter.tick_all(3)

        assert_equal(self.meter.m1.tick.call_args_list, [[], [], []])
        assert_equal(self.meter.m5.tick.call_args_list, [[], [], []])
        assert_equal(self.meter.m15.tick.call_args_list, [[], [], []])
        assert_equal(self.meter.day.tick.call_args_list, [[], [], []])

    @mock.patch('appmetrics.meter.time')
    def test_tick(self, time_mod):
        self.meter.tick_all = mock.Mock()

        time_mod.time.return_value = self.started_on + 2.0
        self.meter.tick()

        time_mod.time.return_value = self.started_on + 8.1
        self.meter.tick()

        time_mod.time.return_value = self.started_on + 18.2
        self.meter.tick()

        assert_equal(self.meter.tick_all.call_args_list, [mock.call(1), mock.call(2)])

    def test_raw_data(self):
        self.meter.count = 5
        assert_equal(self.meter.raw_data(), 5)

    def test_get(self):
        self.meter.tick = mock.Mock()

        expected = dict(
            kind="meter",
            count=0,
            mean=0.0,
            one=0.0,
            five=0.0,
            fifteen=0.0,
            day=0.0
        )

        assert_equal(self.meter.get(), expected)
        assert_equal(self.meter.tick.call_args_list, [[]])

    @mock.patch('appmetrics.meter.time')
    def test_functional(self, time_mod):
        # almost functional test, except for the time module patch trick, needed
        # to make the time pass faster

        time_mod.time.return_value = self.started_on + 0.1
        self.check(0, 0, 0, 0, 0, 0)

        time_mod.time.return_value = self.started_on + 2.5
        self.meter.notify(1)

        time_mod.time.return_value = self.started_on + 3.0
        self.check(1, 1.0 / 3, 0, 0, 0, 0)

        time_mod.time.return_value = self.started_on + 5.1
        self.check(1, 1.0 / 5.1, 0.2, 0.2, 0.2, 0.2)

        self.meter.notify(23)

        time_mod.time.return_value = self.started_on + 60.1
        self.check(
            24, 24.0 / 60.1, 0.23981328001523144, 0.23085723900577299,
            0.2122512344327359, 0.2001387676963957)

        time_mod.time.return_value = self.started_on + (60 * 5) + 0.1
        self.check(
            24, 24.0 / 300.1, 0.004392333437481882, 0.1037308440614258,
            0.16256923530491085, 0.19958359810087625)

    def check(self, count, mean, one, five, fifteen, day):
        data = self.meter.get()

        log.info("data=%s", data)

        assert_equal(data['count'], count)
        assert_almost_equal(data['mean'], mean)
        assert_almost_equal(data['one'], one)
        assert_almost_equal(data['five'], five)
        assert_almost_equal(data['fifteen'], fifteen)
        assert_almost_equal(data['day'], day)


import json
import io

import mock
from nose.tools import (
    assert_equal, assert_false, assert_is_instance, raises, assert_raises,
    assert_regexp_matches)
import werkzeug, werkzeug.test

from .. import wsgi, metrics, py3comp


def env(path, **kwargs):
    data = werkzeug.test.EnvironBuilder(path, environ_overrides=kwargs)
    return data.get_environ()


def req(body):
    data = werkzeug.test.EnvironBuilder(data=json.dumps(body), content_type="application/json")
    return data.get_request(cls=werkzeug.wrappers.Request)


def check_dispatching(mw, url, method, expected):
    urls = mw.url_map.bind_to_environ(env(url, REQUEST_METHOD=method))
    try:
        endpoint, _ = urls.match()
    except expected:
        pass
    except Exception:
        raise
    else:
        assert_equal(endpoint, expected)


def test_dispatching():
    tests = [
        ("/_app-metrics", 'GET', werkzeug.exceptions.NotFound),
        ("/_app-metrics/metrics", 'GET', wsgi.handle_metrics_list),
        ("/_app-metrics/metrics", 'POST', werkzeug.exceptions.MethodNotAllowed),
        ("/_app-metrics/metrics/test", 'GET', wsgi.handle_metric_show),
        ("/_app-metrics/metrics/test", 'PUT', wsgi.handle_metric_new),
        ("/_app-metrics/metrics/test", 'POST', wsgi.handle_metric_update),
        ("/_app-metrics/metrics/test", 'DELETE', wsgi.handle_metric_delete),
        ("/_app-metrics/metrics/test", 'OPTIONS', werkzeug.exceptions.MethodNotAllowed),
        ("/_app-metrics/metrics/test/sub", 'GET', werkzeug.routing.NotFound),
    ]

    mw = wsgi.AppMetricsMiddleware(None)

    for url, method, expected in tests:
        yield lambda url, method: check_dispatching(mw, url, method, expected), url, method


def test_dispatching_root():
    tests = [
        ("/metrics", 'GET', wsgi.handle_metrics_list),
        ("/metrics", 'POST', werkzeug.exceptions.MethodNotAllowed),
        ("/metrics/test", 'GET', wsgi.handle_metric_show),
        ("/metrics/test", 'PUT', wsgi.handle_metric_new),
        ("/metrics/test", 'POST', wsgi.handle_metric_update),
        ("/metrics/test", 'DELETE', wsgi.handle_metric_delete),
        ("/metrics/test", 'OPTIONS', werkzeug.exceptions.MethodNotAllowed),
        ("/metrics/test/sub", 'GET', werkzeug.routing.NotFound),
        ]

    mw = wsgi.AppMetricsMiddleware(None, "")

    for url, method, expected in tests:
        yield lambda url, method: check_dispatching(mw, url, method, expected), url, method


class TestAppMetricsMiddleware(object):
    def setUp(self):
        self.app = mock.Mock()
        self.start_response = mock.Mock()

        self.patch = mock.patch('appmetrics.wsgi.handle_metrics_list')
        self.handler = self.patch.start()
        self.mw = wsgi.AppMetricsMiddleware(self.app)

    def tearDown(self):
        self.patch.stop()

    def test_call_not_matching(self):
        res = self.mw(env("/"), self.start_response)
        assert_equal(res, self.app.return_value)
        assert_false(self.start_response.called)
        assert_equal(
            self.app.call_args_list,
            [mock.call(env("/"), self.start_response)])

    def test_call_not_matching_2(self):
        res = self.mw(env("/test"), self.start_response)
        assert_equal(res, self.app.return_value)
        assert_false(self.start_response.called)
        assert_equal(
            self.app.call_args_list,
            [mock.call(env("/test"), self.start_response)])

    def test_call_not_matching_3(self):
        res = self.mw(env("/_app-metrics/test/sub"), self.start_response)
        assert_equal(res, self.app.return_value)
        assert_false(self.start_response.called)
        assert_equal(
            self.app.call_args_list,
            [mock.call(env("/_app-metrics/test/sub"), self.start_response)])

    def test_call_with_invalid_status(self):
        self.handler.side_effect = ValueError()

        self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='GET'), self.start_response)

        assert_equal(
            self.start_response.call_args_list,
            [mock.call("500 INTERNAL SERVER ERROR", mock.ANY)]
        )

    def test_call_with_error_implicit(self):
        self.handler.side_effect = werkzeug.exceptions.BadRequest()

        body = self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='GET'), self.start_response)

        expected_body = json.dumps(werkzeug.exceptions.BadRequest.description)
        assert_equal(b"".join(body), expected_body.encode('utf8'))

        expected_headers = [
            ('Content-Type', 'application/json'),
            ('Content-Length', str(len(expected_body)))
        ]
        assert_equal(
            self.start_response.call_args_list,
            [mock.call("400 BAD REQUEST", expected_headers)]
        )

    def test_call_with_error_explicit(self):
        self.handler.side_effect = werkzeug.exceptions.BadRequest(description="bad request received")

        body = self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='GET'), self.start_response)

        expected_body = json.dumps("bad request received")

        assert_equal(b"".join(body), expected_body.encode('utf8'))

        expected_headers = [
            ('Content-Type', 'application/json'),
            ('Content-Length', str(len(expected_body)))
        ]
        assert_equal(
            self.start_response.call_args_list,
            [mock.call("400 BAD REQUEST", expected_headers)]
        )

    def test_call_with_invalid_method(self):
        self.handler.side_effect = werkzeug.exceptions.BadRequest()

        body = self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='POST'), self.start_response)

        expected_body = json.dumps(werkzeug.exceptions.MethodNotAllowed.description)
        assert_equal(b"".join(body), expected_body.encode('utf8'))

        assert_equal(
            self.start_response.call_args_list,
            [mock.call("405 METHOD NOT ALLOWED", mock.ANY)]
        )

        headers = dict(self.start_response.call_args_list[0][0][1])
        assert_equal(headers['Content-Type'], "application/json")
        assert_equal(headers['Content-Length'], str(len(expected_body)))
        allow = {x.strip() for x in headers['Allow'].split(",")}
        assert_equal(allow, {"HEAD", "GET"})


    def test_call_ok(self):
        self.handler.return_value = json.dumps("results")

        body = self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='GET'), self.start_response)

        expected_headers = [
            ('Content-Type', 'application/json'),
            ('Content-Length', str(len(self.handler.return_value)))
        ]
        assert_equal(
            self.start_response.call_args_list,
            [mock.call("200 OK", expected_headers)]
        )

        expected = json.dumps("results")

        assert_equal(b"".join(body), expected.encode('utf8'))


    def test_call_with_unicode(self):
        if py3comp.PY3:
            self.handler.return_value = json.dumps("results")
        else:
            self.handler.return_value = json.dumps("results").decode('utf8')

        body = self.mw(env("/_app-metrics/metrics", REQUEST_METHOD='GET'), self.start_response)

        expected_body = json.dumps("results")
        assert_equal(b"".join(body), expected_body.encode('utf8'))

        expected_headers = [
            ('Content-Type', 'application/json'),
            ('Content-Length', str(len(expected_body)))
        ]
        assert_equal(
            self.start_response.call_args_list,
            [mock.call("200 OK", expected_headers)]
        )


class TestWSGIHandlers(object):
    def setUp(self):
        self.original_registry = metrics.REGISTRY
        metrics.REGISTRY.clear()

        self.original_tags = metrics.TAGS
        metrics.TAGS.clear()

    def tearDown(self):
        metrics.REGISTRY.clear()
        metrics.REGISTRY.update(self.original_registry)

        metrics.TAGS.clear()
        metrics.TAGS.update(self.original_tags)

    @mock.patch('appmetrics.wsgi.metrics.metrics')
    def test_handle_metrics_list(self, metrics):
        metrics.return_value = ["test1", "test2"]

        assert_equal(wsgi.handle_metrics_list(mock.Mock()), '["test1", "test2"]')

    @mock.patch('appmetrics.wsgi.metrics.metric')
    def test_handle_metric_show(self, metric):
        metric().get.return_value = "this is a test"

        assert_equal(wsgi.handle_metric_show(mock.Mock(), "test"), '"this is a test"')

    @mock.patch('appmetrics.wsgi.metrics.metric')
    def test_handle_metric_show_not_found(self, metric):
        metric.side_effect = KeyError("key")

        with assert_raises(werkzeug.exceptions.NotFound) as exc:
            wsgi.handle_metric_show(mock.Mock(), "test")

        assert_equal(exc.exception.description, "No such metric: 'test'")

    def test_handle_metric_delete(self):
        with mock.patch.dict('appmetrics.metrics.REGISTRY', dict(test=mock.Mock())):
            res = wsgi.handle_metric_delete(mock.Mock(), "test")

            assert_equal(res, "deleted")
            assert_equal(metrics.REGISTRY, dict())

    def test_handle_metric_delete_not_found(self):
        with mock.patch.dict('appmetrics.metrics.REGISTRY', dict(none="test")):
            res = wsgi.handle_metric_delete(mock.Mock(), "test")

            assert_equal(res, "not deleted")
            assert_equal(metrics.REGISTRY, dict(none="test"))

    @mock.patch('appmetrics.wsgi.metrics.tags')
    def test_handle_tags_list(self, tags):
        tags.return_value = dict(tag1=["test1", "test2"], tag2=["test3"])

        assert_equal(wsgi.handle_tags_list(mock.Mock()), '["tag1", "tag2"]')

    def test_handle_tag_add(self):
        metrics.REGISTRY["test1"] = mock.Mock()

        res = wsgi.handle_tag_add(mock.Mock(), "tag1", "test1")

        assert_equal(res, "")
        assert_equal(metrics.TAGS, {"tag1": {"test1"}})

    @raises(werkzeug.exceptions.BadRequest)
    def test_handle_tag_add_invalid(self):
        res = wsgi.handle_tag_add(mock.Mock(), "tag1", "test1")

        assert_equal(res, "")
        assert_equal(metrics.TAGS, {"tag1": {"test1"}})

    def test_handle_untag_not_existing(self):
        res = wsgi.handle_untag(mock.Mock(), "tag1", "test1")
        assert_equal(res, "not deleted")

    def test_handle_untag(self):
        metrics.TAGS["tag1"] = {"test1"}

        res = wsgi.handle_untag(mock.Mock(), "tag1", "test1")
        assert_equal(res, "deleted")

    @raises(werkzeug.exceptions.NotFound)
    def test_handle_tag_show_not_found(self):
        wsgi.handle_tag_show(mock.Mock(), "tag1")

    def test_handle_tag_show(self):
        metrics.new_histogram("test1")
        metrics.tag("test1", "tag1")

        res = wsgi.handle_tag_show(mock.Mock(), "tag1")
        assert_equal(res, '["test1"]')

    def test_handle_tag_show_no_expand(self):
        metrics.new_histogram("test1")
        metrics.tag("test1", "tag1")

        res = wsgi.handle_tag_show(mock.Mock(args={"expand": 'false'}), "tag1")
        assert_equal(res, '["test1"]')

    @mock.patch('appmetrics.metrics.metrics_by_tag')
    def test_handle_tag_show_no_expand(self, mbt):
        mbt.return_value = "this is a test"
        metrics.new_histogram("test1")
        metrics.tag("test1", "tag1")

        res = wsgi.handle_tag_show(mock.Mock(args={"expand": 'true'}), "tag1")
        assert_equal(res, '"this is a test"')

    @raises(werkzeug.exceptions.UnsupportedMediaType)
    def test_get_body_no_content_type(self):
        request = werkzeug.wrappers.Request(dict(CONTENT_LENGTH=10))

        wsgi.get_body(request)

    @raises(werkzeug.exceptions.UnsupportedMediaType)
    def test_get_body_bad_content_type(self):
        request = werkzeug.wrappers.Request(dict(CONTENT_LENGTH=10, CONTENT_TYPE='text/html'))

        wsgi.get_body(request)

    def test_get_body_bad_content(self):
        env = {'CONTENT_LENGTH': 4, 'CONTENT_TYPE': "application/json", 'wsgi.input': io.StringIO(u"test wrong")}
        request = werkzeug.wrappers.Request(env)

        with assert_raises(werkzeug.exceptions.BadRequest) as exc:
            wsgi.get_body(request)
        assert_equal(exc.exception.description, "invalid json")

    def test_get_body(self):
        env = {'CONTENT_LENGTH': 6, 'CONTENT_TYPE': "application/json", 'wsgi.input': io.StringIO(u'"test" with garbage')}
        request = werkzeug.wrappers.Request(env)

        assert_equal(wsgi.get_body(request), 'test')

    def test_handle_metric_new_missing_type(self):
        with assert_raises(werkzeug.exceptions.BadRequest) as exc:
            wsgi.handle_metric_new(req(dict()), "test")
        assert_equal(exc.exception.description, "metric type not provided")

    def test_handle_metric_new_invalid_type(self):
        with assert_raises(werkzeug.exceptions.BadRequest) as exc:
            wsgi.handle_metric_new(req(dict(type="xxx")), "test")
        assert_regexp_matches(exc.exception.description, "invalid metric type: .*'xxx'")

    def test_handle_metric_new_app_error(self):
        wsgi.handle_metric_new(req(dict(type="gauge")), "test")

        with assert_raises(werkzeug.exceptions.BadRequest) as exc:
            wsgi.handle_metric_new(req(dict(type="gauge")), "test")
        assert_equal(exc.exception.description, "can't create metric gauge('test'): Metric test already exists of type Gauge")

    def test_handle_metric_new_generic_error(self):
        new_gauge = mock.Mock(side_effect=ValueError("an error"))

        with mock.patch.dict('appmetrics.wsgi.metrics.METRIC_TYPES', gauge=new_gauge):
            with assert_raises(werkzeug.exceptions.BadRequest) as exc:
                wsgi.handle_metric_new(req(dict(type="gauge")), "test")
            assert_equal(exc.exception.description, "can't create metric gauge('test')")

    def test_handle_metric_new_metric(self):
        res = wsgi.handle_metric_new(req(dict(type="gauge")), "test")
        assert_equal(res, "")

        metric = metrics.metric("test")
        assert_is_instance(metric, metrics.simple_metrics.Gauge)

    def test_handle_metric_update_missing_value(self):

        with assert_raises(werkzeug.exceptions.BadRequest) as exc:
            wsgi.handle_metric_update(req(dict()), "test")
        assert_equal(exc.exception.description, "metric value not provided")

    @raises(werkzeug.exceptions.NotFound)
    def test_handle_metric_update_missing_metric(self):
        wsgi.handle_metric_update(req(dict(value=1)), "test")

    def test_handle_metric_update(self):
        metric = metrics.new_gauge("test")

        res = wsgi.handle_metric_update(req(dict(value=1.5)), "test")
        assert_equal(res, "")

        assert_equal(metric.get(), dict(kind="gauge", value=1.5))




##  Module py3comp.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.

"""
Python 3 compatibility
"""

import sys
import json

PY3 = sys.version_info[0] == 3

if PY3:
    xrange = range

    def iteritems(d, **kw):
        return d.items(**kw)

    __builtin_zip = zip

    zip = lambda *args: list(__builtin_zip(*args))
else:
    xrange = xrange

    def iteritems(d, **kw):
        return d.iteritems(**kw)

    zip = zip


def assert_items_equal(*args):
    from nose import tools as nt

    if hasattr(nt, 'assert_items_equal'):
        return nt.assert_items_equal(*args)
    else:
        return nt.assert_equal(set(args[0]), set(args[1]))


def json_load(stream, charset):
    # thanks very much, python 3...
    if PY3:
        raw_data = stream.read()
        if isinstance(raw_data, str):
            return json.loads(raw_data)
        else:
            return json.loads(raw_data.decode(charset))
    else:
        return json.load(stream)




##  Module wsgi.py
##
##  Copyright (c) 2014 Antonio Valente <y3sman@gmail.com>
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##  http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.


"""
WSGI middleware
"""

import logging, logging.config
import json

import werkzeug

from . import metrics, exceptions, py3comp


log = logging.getLogger("appmetrics.wsgi")


class AppMetricsMiddleware(object):
    """
    WSGI middleware for AppMetrics

    Usage:

    Instantiate me with the wrapped WSGI application. This middleware looks for request paths starting
    with "/_app-metrics": if not found, the wrapped application is called. The following resources are defined:

    ``/_app-metrics/metrics``
      - **GET**: return the list of the registered metrics
    ``/_app-metrics/metrics/<name>``
      - **GET**: return the value of the given metric or ``404``.
      - **PUT**: create a new metric with the given name. The body must be a ``JSON`` object with a
        mandatory attribute named ``"type"`` which must be one of the metrics types allowed,
        by the ``"metrics.METRIC_TYPES"`` dictionary, while the other attributes are
        passed to the ``new_<type>`` function as keyword arguments.
        Request's ``content-type`` must be ``"application/json"``.
      - **POST**: add a new value to the metric. The body must be a ``JSON`` object with a mandatory
        attribute named ``"value"``: the notify method will be called with the given value.
        Other attributes are ignored.
        Request's ``content-type`` must be ``"application/json"``.
      - **DELETE**: remove the metric with the given name. Return "deleted" or "not deleted".
    ``/_app-metrics/tags``
      - **GET**: return the list of registered tags
    ``/_app-metrics/tags/<name>``
      - **GET**: return the metrics tagged with the given tag. If the value of the ``GET`` parameter ``"expand"``
        is ``"true"``, a JSON object is returned, with the name of each tagged metric as keys and corresponding values.
        If it is ``"false"`` or not provided, the list of metric names is returned.
        Return a ``404`` if the tag does not exist
    ``/_app-metrics/tags/<tag_name>/<metric_name>``
      - **PUT**: tag the metric named ``<metric_name>`` with ``<tag_name>``. Return a ``400`` if the given metric
        does not exist.
      - **DELETE**: remove the tag ``<tag_name>`` from ``<metric_name>``. Return "deleted" or "not deleted". If
        ``<tag_name>`` is no longer used, it gets implicitly removed.


    The root can be different from "/_app-metrics", you can set it on middleware constructor.
    """

    def __init__(self, app, root="_app-metrics", extra_headers=None, mimetype="application/json"):
        """
        parameters:
        - app: wrapped WSGI application
        - root: path root to look for
        - extra_headers: extra headers that will be appended to the return headers
        """
        self.app = app
        self.root = "/" + root.strip("/").strip()
        self.extra_headers = extra_headers or {}
        self.mimetype = mimetype

        self.url_map = werkzeug.routing.Map([
            werkzeug.routing.Submount(self.root, [
                werkzeug.routing.Rule("/metrics", endpoint=handle_metrics_list, methods=['GET']),
                werkzeug.routing.Rule("/metrics/<name>", endpoint=handle_metric_show, methods=['GET']),
                werkzeug.routing.Rule("/metrics/<name>", endpoint=handle_metric_new, methods=['PUT']),
                werkzeug.routing.Rule("/metrics/<name>", endpoint=handle_metric_update, methods=['POST']),
                werkzeug.routing.Rule("/metrics/<name>", endpoint=handle_metric_delete, methods=['DELETE']),
                werkzeug.routing.Rule("/tags", endpoint=handle_tags_list, methods=['GET']),
                werkzeug.routing.Rule("/tags/<tag_name>", endpoint=handle_tag_show, methods=['GET']),
                werkzeug.routing.Rule("/tags/<tag_name>/<metric_name>", endpoint=handle_tag_add, methods=['PUT']),
                werkzeug.routing.Rule("/tags/<tag_name>/<metric_name>", endpoint=handle_untag, methods=['DELETE']),
            ])
        ])

    def get_response(self, body, code, headers=None):
        if headers is None:
            headers = []
        headers = dict(headers).copy()
        headers.update(self.extra_headers)
        return werkzeug.wrappers.Response(body, code, headers.items(), self.mimetype)

    def jsonize_error(self, exception, environ):
        return self.get_response(json.dumps(exception.description), exception.code, exception.get_headers(environ))

    def __call__(self, environ, start_response):
        """WSGI application interface"""

        urls = self.url_map.bind_to_environ(environ)

        try:
            endpoint, args = urls.match()
        except werkzeug.exceptions.NotFound:
            # the request did not match, go on with wsgi stack
            return self.app(environ, start_response)

        except werkzeug.exceptions.HTTPException as e:
            response = self.jsonize_error(e, environ)

        else:
            request = werkzeug.wrappers.Request(environ, populate_request=False)
            try:
                body = endpoint(request, **args)
                response = self.get_response(body, 200)
            except werkzeug.exceptions.HTTPException as e:
                response = self.jsonize_error(e, environ)

            except Exception as e:
                log.debug("Unhandled exception: %s", e, exc_info=True)
                response = self.get_response("Internal Server Error", 500)

        return response(environ, start_response)


def get_body(request):
    # get content type
    ctype = request.mimetype
    if not ctype or ctype != "application/json":
        raise werkzeug.exceptions.UnsupportedMediaType()

    # get content data
    try:
        return py3comp.json_load(request.stream, request.charset)
    except ValueError as e:
        log.debug("Invalid body: %s", e)
        raise werkzeug.exceptions.BadRequest(description="invalid json")


def handle_metrics_list(request):
    return json.dumps(metrics.metrics())


def handle_metric_show(request, name):
    try:
        metric = metrics.metric(name)
    except KeyError:
        raise werkzeug.exceptions.NotFound(("No such metric: {!r}".format(name)))

    return json.dumps(metric.get())


def handle_metric_delete(request, name):
    res = metrics.delete_metric(name)

    return "deleted" if res else "not deleted"


def handle_metric_new(request, name):
    data = get_body(request)

    type_ = data.pop('type', None)
    if not type_:
        raise werkzeug.exceptions.BadRequest(description="metric type not provided")

    metric_type = metrics.METRIC_TYPES.get(type_)

    if not metric_type:
        raise werkzeug.exceptions.BadRequest("invalid metric type: {!r}".format(type_))

    try:
        metric_type(name, **data)
    except exceptions.AppMetricsError as e:
        raise werkzeug.exceptions.BadRequest("can't create metric {}({!r}): {}".format(type_, name, e))
    except Exception as e:
        log.debug(str(e), exc_info=True)
        raise werkzeug.exceptions.BadRequest("can't create metric {}({!r})".format(type_, name))

    return ""


def handle_metric_update(request, name):
    data = get_body(request)

    value = data.pop('value', None)
    if value is None:
        raise werkzeug.exceptions.BadRequest("metric value not provided")

    try:
        metric = metrics.metric(name)
    except KeyError:
        raise werkzeug.exceptions.NotFound()

    metric.notify(value)

    return ""


def handle_tags_list(request):
    return json.dumps(sorted(metrics.tags().keys()))


def handle_tag_show(request, tag_name):
    all_tags = metrics.tags()
    if tag_name not in all_tags:
        raise werkzeug.exceptions.NotFound(description="no such tag: {!r}".format(tag_name))

    if request.args.get('expand', 'false') == 'true':
        return json.dumps(metrics.metrics_by_tag(tag_name))
    else:
        return json.dumps(sorted(all_tags[tag_name]))


def handle_tag_add(request, tag_name, metric_name):
    try:
        metrics.tag(metric_name, tag_name)
    except metrics.InvalidMetricError as e:
        raise werkzeug.exceptions.BadRequest(description=str(e))

    return ""


def handle_untag(request, tag_name, metric_name):
    res = metrics.untag(metric_name, tag_name)

    return "deleted" if res else "not deleted"


# useful to run standalone with werkzeug's server:
# $ python -m werkzeug.serving appmetrics.wsgi.standalone_app
# * Running on http://127.0.0.1:5000/

standalone_app = AppMetricsMiddleware(werkzeug.exceptions.NotFound(), "")
